#!/usr/bin/perl

#use Data::Dump qw(dump);
#use Strict;
#use Warnings;

# perl antispam smtp
# relays good mail to another mh
# rejects executable attachments
# uses whitelist and bayesian test
# also automatically maintains whitelist and sample databases
# addresses on redlist can never go on the whitelist
# (c) John Hanna 2003 under the terms of the GPL

use bytes; # get rid of anoying 'Malformed UTF-8' messages

$version='0.3.1';

# load from command line if specified
if($ARGV[0]) {
 $base=$ARGV[0];
} else {
 # the last one is the one used if all else fail
 for ('.','assp','/usr/local/assp','/home/assp','/usr/assp','/assp','.') {
  $base=$_;
  last if -e "$base/assp.cfg";
 }
}

$OutgoingBufSize=102400; # if you have a slow internet connection change this to 4096

loadConfig();
sub loadConfig {
 print "loading config -- base='$base'\n";
 @Config=(
 [0,0,0,heading,'Network Setup'],
 # except for the heading lines, all config lines have the following:
 #  $name,$nicename,$size,$func,$default,$valid,$onchange,$description
 # name is the variable name that holds the data
 # nicename is a human readable pretty display name (oh how nice!)
 # size is the appropriate input box size
 # func is a function called to render the config item
 # default is the default value
 # valid is a regular expression used to clean and validate the input -- no match is an error and $1 is the desired result
 # onchange is a function to be called when this value is changed -- usually undef; just updating the value is enough
 # description is text displayed to help the user figure what to put in the entry

 [smtpDestination,'SMTP Destination',20,textinput,'127.0.0.1:225','(\S*)',undef,
  'The address:port of your message handling system\'s smtp server. For example: 127.0.0.1:125'],
 [AsAService,'As a Service',0,checkbox,'','(\S*)',undef,
  'In Windows 2000 / NT you can run it as a service; requires <a href=http://www.roth.net/perl/Daemon/ target=_blank>win32::daemon</a>. Requires start from the service control panel.'],
 [AsADaemon,'As a Daemon',0,checkbox,'','(\S*)',undef,
  'In Linux/BSD/OsX fork and close file handles, kinda like "perl assp.pl &" but better. Requires restart.'],
 [myName,'My Name',20,textinput,'ASSP-nospam','(\S+)',undef,
  'What the program calls itself in the email "received by" header. Usually ASSP-nospam.'],
 [listenPort,'Listen Port',20,textinput,'125','(\S+)',ConfigChangeMailPort,
  'On what port should ASSP accept smtp connections? Normally 25. You can supply an interface:port to limit connections.'],
 [webAdminPort,'Web Admin Port',20,textinput,55555,'(\S+)',ConfigChangeAdminPort,
  'On what port should ASSP listen for http connections for the web administration interface? If you change this you will need to change the URL on your browser to reconnect. You can supply an interface:port to limit connections.'],
 # I hate password input, but if you like it, uncomment this line and comment the next one. -- just quit bugging me about it!
 #[webAdminPassword,'Web Admin Password',2,passinput,'nospam4me','(.{5,})',undef,
 [webAdminPassword,'Web Admin Password',2,textinput,'nospam4me','(.{5,})',undef,
  'This is your password for the administrative interface -- if you leave this default then you deserve what you get. Must be at least 5 characters long.'],
 [allowAdminConnectionsFrom,'Allow Admin Connections From',40,textinput,'','(\S*)',ConfigMakeRe,
  'This is an optional list of IP addresses from which you will accept web admin connections, seperated by bangs (|). For example: 127.0.0.1|10.
  Blank means accept all connections. 127.0.0.1 means accept connections from only the localhost. Note that IP source
  addresses are very easy to spoof, so this should not be considered as a security feature.
  <font color=red>Also note that if you make a mistake here you will likely disable your
  web admin interface and have to manually edit your configuration file to fix it.</font>'],

 [0,0,0,heading,'Relaying'],
 [acceptAllMail,'Accept All Mail',40,textinput,'127.0.0.|10.|169.254.|172.16.|192.168.','(\S*)',ConfigMakeRe,
  'Denies relaying for hosts besides these. These hosts also contribute to the whitelist.
  For example: 127.0.0.1|10.|169.254.|172.16.|192.168.'],
 [relayHostFile,'Relay Host File',40,textinput,'','(.*)',undef,
  'Like Accept All Mail, but this is a file that contains a list of ip addresses (one per line)
   for whom you want to relay mail. This is an ABSOLUTE path, not relative to base.
   For example: /usr/local/assp/relayhosts'],
 [localDomains,'Local Domains',80,textinput,'putYourDomains.com|here.org','(\S*)',ConfigMakeRe,
  'Addresses in these domains are considered local delivery. Seperate addresses with |. 
  Include all subdomains. For example: put.YourDomains.com|here.org'],
 [localDomainsFile,'Local Domains File',40,textinput,'','(.*)',undef,
  'Like Local Domains, but this is a file that contains a list of host names (one per line)
   for whom you want accept mail. This is an ABSOLUTE path, not relative to base.
   For example: /usr/local/assp/locals'],
 [relayHost,'Relay Host',40,textinput,'','(\S*)',undef,
  'Your isp\'s mail relayhost (smarthost). For example: mail.isp.com:25<br>
  If you run Exchange or Notes and you want assp to automatically update the nonspam database and
  the whitelist, then use your isp\'s smtp relay host here. Blank means no relayhost.
  You only need this if your clients don\'t deliver mail through SMTP.'],
 [relayPort,'Relay Port',20,textinput,'','(\S*)',ConfigChangeRelayPort,
  'Tell your mail server to connect to this port (eg 127.0.0.1:225 or 127.0.0.2) as it\'s smarthost / relayhost. For example: 225
  Note that you\'ll want to keep the relayPort protected from external access by your firewall or else
  spammers can use it to send as much spam as they want. You can supply an interface:port to limit connections.'],

 [0,0,0,heading,'SPAM Control'],
 [BlockExes,'Block Executable Attachments',0,checkbox,1,'(.*)',undef,
  'Most executable attachments are email viruses. Check this box to not allow delivery of mail
  with executable attachments from non-whitelisted users.'],
 [BlockAllExes,'Block Whitelisted Exe Attachments too',0,checkbox,1,'(.*)',undef,
  'Check the above box AND this box to block attachments even from whitelisted & local senders.'],
 [spamaddresses,'Spam Addresses',60,textinput,'put|your|spambucket|addresses|here','(\S*)',ConfigMakeRe,
  'Mail to any of these users at any of the localDomains are always spam unless from someone on the whitelist; don\'t include the @domain.com part.'],
 [spamLovers,'Spam-Lover Addresses',60,textinput,'','(\S*)',ConfigMakeRe,
  'Mail solely to any of these users at any of the localDomains are never blocked, which is to say they get all their spam; mail to more than one recipient is blocked unless all recipients are spamLovers.
  Separate addresses with bangs: | For example: postmaster|julie_vipul'],
 [noProcessing,'Unprocessed Addresses',40,textinput,'','(\S*)',ConfigMakeRe,
  'Mail solely to or from any of these users at any of the localDomains are ignored by ASSP, like
  a more effecient version of spamLovers & redlist combined.
  Separate addresses with bangs: | For example: postmaster|julie_vipul'],
 [whiteListedDomains,'Whitelisted Domains',60,textinput,'sourceforge.net','(\S*)',ConfigMakeRe,
  'Domains from which you want to receive all mail -- your ISP, domain registeration, mail list servers, stock broker, or other key business partners might be good candidates.
  Note this matches the end of the address, so if you don\'t want to match subdomains then
  include the @. For example: sourceforge.net|@google.com'],
 [blackListedDomains,'Blacklisted Domains',60,textinput,'','(\S*)',ConfigMakeRe,
  'Domains from which you always want to reject mail -- they only send you spam. For example: email-deliveries.net|xxxpics.com'],
 [whiteRe,'Expression to Identify Non-Spam',60,textinput,'','(.*)',undef,
  'If an incoming email matches this Perl regular expression it will be considered non-spam. For example:
  Secret Ham Password|303\D{0,3}730\D{0,3}4[12]\d\d<br>For help writing regular expressions click
  <a href="http://www.perlmonks.org/index.pl?node=perlre" target=_blank>here</a>. Note that flags are "si" and
  the header as well as body is scanned. Some things you might include here are your office phone
  number or street address -- spam rarely includes these details.'],
 [blackRe,'Expression to Identify Spam',60,textinput,'http://\\w+@','(.*)',undef,
  'If an incoming email that\'s not local or whitelisted matches this Perl regular expression
  it will be considered spam. May match text from the body or header of the email.
  For example: penis|virgin|X-Priority: 1'],
 [redRe,'Expression to Identify Redlisted Mail',60,textinput,'','(.*)',undef,
  'If an email header matches this Perl regular expression it will be considered redlisted.
  For example: [autoreply]'],
 [WhitelistOnly,'Reject All But Whitelisted Mail',0,checkbox,0,'(.*)',undef,
  'Check this if you don\'t want Bayesian filtering and want to reject all mail from
   anyone not whitelisted. Note: this turns the redlist into a blacklist.'],
 [NoMaillog,'Don\'t log mail',0,checkbox,0,'(.*)',undef,
  'Check this if you\'re using Whitelist-Only and don\'t care to save mail to build
   the Bayesian database.'],
 [NoGoodhosts,'Disable Good Hosts Antispam',0,checkbox,0,'(.*)',undef,
  'Check this to disable the Good-Hosts antispam feature.'],
 [MaxWhitelistDays,'Max Whitelist Days',5,textinput,'400','(\d+)',undef,
  'This is the number of days an address will be kept on the whitelist without receiving email.'],
 [AddSpamProbHeadder,'Add Spam Probability Header?',0,checkbox,1,'(.*)',undef,
  'Adds a line to the email header "X-Assp-Spam-Prob: 0.0123" Probs range from 0 to +1 where > 0.6 = spam.'],
 [NoExternalSpamProb,'Block Outgoing Spam-Prob header?',0,checkbox,0,'(.*)',undef,
  'Check this box if you don\'t want your X-Assp-Spam-Prob header on external mail -- Note this
   means mail from local users to local users will also be missing the header.'],
 [AddSpamHeadder,'Add Spam Header?',0,checkbox,1,'(.*)',undef,
  'Adds a line to the email header "X-Assp-Spam: YES" if the message is spam.'],
 [UseSubjectsAsMaillogNames,'Use Subject as Maillog Names',0,checkbox,'','(.*)',undef,
  'You can turn this on to help you manually identify mail in your spam and non-spam collections.'],
 [TestMode,'Test Mode',0,checkbox,1,'(.*)',undef,
  'If TestMode is true all bayesian-spam is delivered.'],
 [spamSubject,'Prepend Spam Subject',10,textinput,'[SPAM]','(.*)',undef,
  'If TestMode and message is spam spamSubject gets prepended to the subject of the email. For example; [SPAM]'],

 [0,0,0,heading,'Email Interface'],
 [EmailInterfaceOk,'Enable Email Interface',0,checkbox,1,'(.*)',undef,
  'Checked means that you want to ASSP to intercept and parse mail to the following addresses
   at any of your localdomains. If you are using RelayHost and RelayPort see
   <a href=http://assp.sourceforge.net/fom/cache/70.html target=_blank>this note</a>.'],
 [EmailSpam,'Report Spam Address',20,textinput,'assp-spam','(.*)',undef,
  'Any mail sent by local or authenticated users to this address will be interpreted as a
   spam report. No mail is delivered! For example: assp-spam'],
 [EmailHam,'Report not-Spam Address',20,textinput,'assp-notspam','(.*)',undef,
  'Any mail sent by local or authenticated users to this address will be interpreted as a
   false-positive (not-spam) report. No mail is delivered! For example: assp-notspam'],
 [EmailWhitelist,'Add to Whitelist Address',20,textinput,'assp-white','(.*)',undef,
  'Any mail sent by local or authenticated users to this address will be interpreted as a request
   to add addresses to the whitelist. No mail is delivered! For example: assp-white'],
 [EmailFrom,'From Address for Email',20,textinput,'ASSP <>','(.+)',undef,
  'Email sent from ASSP acknowledging your submissions will be sent from this address. Some
   mailers don\'t like the default setting. For example: ASSP <> or Mail Administrator
   <mailadmin@mydomain.com>'],

 [0,0,0,heading,'File Paths'],
 ['base','Directory Base',40,textinput,'.','             ',undef,
  'All paths are relative to this folder. <b>Note: this must be changed as a command line parameter and is displayed here for reference only.</b>'],
 [spamlog,'Spam Collection',40,textinput,'spam','(\S+)',undef,
  'The folder to save the collection of spam emails. For example: spam'],
 [notspamlog,'Not-spam Collection',40,textinput,'notspam','(\S+)',undef,
  'The folder to save the collection of not-spam emails. For example: notspam'],
 [incomingOkMail,'External mail that wasn\'t spam (mostly)',40,textinput,'','(\S*)',undef,
  'The folder to save bayesian non-spam and viri. Leave this blank to not save
   these files (default). Most mail admins don\'t want these files, but if you want to
   keep copies of all incoming mail then put in a directory name. Note: you must create
   the directory. For example: other'],
 [correctedspam,'False-negative Collection',40,textinput,'errors/spam','(\S+)',undef,
  'Spam that got through -- counts double. For example: errors/spam'],
 [correctednotspam,'False-positive Collection',40,textinput,'errors/notspam','(\S+)',undef,
  'Good mail that was listed as spam, count 4x. For example: errors/notspam'],
 [spamdb,'Spam Bayesian Database File',40,textinput,'spamdb','(\S+)',undef,
  'The output file from rebuildspamdb.pl.'],
 [whitelistdb,'Email Whitelist Database File',40,textinput,'whitelist','(\S+)',undef,
  'The file with the whitelist.'],
 [redlistdb,'Email Redlist Database File',40,textinput,'redlist','(\S+)',undef,
  'The file with the redlist.'],
 [dnsbl,'DNS Blacklist Database File',40,textinput,'dnsbl','(\S*)',undef,
  'The file with the current DNSBL -- make this blank if you don\'t use it.'],
 [logfile,'ASSP Logfile',40,textinput,'maillog.txt','(\S*)',ConfigChangeLogfile,
  'Blank if you don\'t want a log file.'],
 [pidfile,'PID File',40,textinput,'pid','(\S*)',undef,
  'Blank to skip writing a pid file. *nix users need pid files. Leave it blank in Windows.
   You have to restart the service before you get a pid file in the new location.'],

 [0,0,0,heading,'Logging'],
 [silent,'Silent Mode',0,checkbox,'','(.*)',undef,
  'Checked means don\'t print log messages to the console. AsADaemon overrides this.'],
 [DEBUG,'Debug Mode',0,checkbox,'','(.*)',ConfigDEBUG,
  'Checked means you want debugging to go to a .dbg file. Leave this unchecked unless there is a program error you are trying to track down.'],
 [sendNoopInfo,'Send NOOP Info',0,checkbox,'','(.*)',undef,
  'Checked means you want ASSP to send a "NOOP Connection from $ip" message to your SMTP server. (Postfix croaks on this.)'],
 [totalizeSpamStats,'Upload Colsolated Spam Statistics',0,checkbox,1,'(.*)',undef,
  'Checked means your ASSP will upload it\'s totalled statistics to the <a href=http://assp.sourceforge.net/cgi-bin/total.pl target=_blank>ASSP web site totals</a>.
  This is a great marketing tool for the ASSP project, please don\'t disable it unless you\'ve got
  a good reason to. No private information is being disclosed by this upload.'],

 [0,0,0,heading,'Security'],
 [runAsUser,'Run as UID',20,textinput,'','(\S*)',undef,
  'The *nix user name to assume after startup: assp or nobody -- requires ASSP restart.'],
 [runAsGroup,'Run as GID',20,textinput,'','(\S*)',undef,
  'The *nix group to assume after startup: assp or nogroup -- requires ASSP restart.'],
 [ChangeRoot,'Change Root',60,textinput,'','(.*)',undef,
  'Non-blank means to run in chroot jail in *nix. You need an etc/protocols file to make this work -- copy or link the file to your new root directory.  -- requires ASSP restart.'],

 [0,0,0,heading,'Other Settings'],
 [MaxFiles,'Max Files',10,textinput,'14003','(\d+)',undef,
  'If you\'re not using subjects as file names, MaxFiles is the maximum
  number of files to keep in each collection (spam and nonspam) -- it\'s actually
  less than this -- files get a random number between 1 and $MaxFiles.'],
 [MaxErrors,'Max Errors',10,textinput,'10','(\d+)',undef,
  'If the smtp destination sends $MaxErrors 501s, 502s, 503s... the connection is dropped.'],
 [RestartEvery,'Restart Every',10,textinput,'0','(\d+)',undef,
  'Program terminates after this many seconds -- this is really only useful if ASSP runs in a script that restarts it after
   it stops. Note: the current timeout must expire before the new setting is loaded.'],
 [UpdateWhitelist,'Save Whitelist',10,textinput,'3600','(\d+)',undef,
  'Save a copy of the white list every this many seconds. Note: the current timeout must expire
   before the new setting is loaded, or you can restart.'],
 [NoRelaying,'No Relaying Error',80,textinput,'550 Relaying not allowed','(5\d\d .*)',undef,
  'SMTP error message to deny relaying.'],
 [AttachmentError,'Executable Attachment Error',80,textinput,'500 Executable attachments are not allowed -- Compress before mailing.','(5\d\d .*)',undef,
  'SMTP error message to reject executable attachments.'],
 [SpamError,'Spam Error',80,textinput,'500 Mail appears to be unsolicited -- send error reports to postmaster@yourdomain.com','(5\d\d .*)',undef,
  'SMTP error message to reject spam.'],
 );

 # allow override for default web admin port
 if($ARGV[1]=~/^\d+$/) {
  for (@Config) {
   if($_->[0] eq 'webAdminPort' ) {
    $_->[4]=$ARGV[1];
    last;
   }
  }
 }
 open(F,"<$base/assp.cfg"); local $/; (%Config)=split(/:=|\n/,<F>); close F;
 # load defaults
 unless( %Config ) { for $c (@Config) {$Config{$c->[0]}=$c->[4] if $c->[0];} }

 $Color1='bgcolor=#DED2CA';
 $Color2='bgcolor=#F1EAE7';
 $Color3='bgcolor=#D4C3B2';
 #$Color3='bgcolor=#0A8B71';
 $Color4='bgcolor=#C5AD94';

 # regular expression to find attachments to be rejected
 # if you don't know perl regular expressions, leave this alone
 # if you don't want to reject attachments then comment out the next four lines
 # Note that whitelisted senders can send executable attachments.
 $badattach=&re(<<'EOT');
content-\w+: .*\s+.*name\s*=\s*".*\.(exe|scr|pif|vb[es]|js|jse|ws[fh]|sh[sb]|lnk|bat|com|ht[ab])"
content-\w+: .*\s+.*name\s*=\s*.*\.(exe|scr|pif|vb[es]|js|jse|ws[fh]|sh[sb]|lnk|bat|com|ht[ab])\s
EOT

 fixConfigSettings();
 $SIG{HUP}=\&reloadConfigFile;
}

# Notes on general operation & program structure
# I'm using IO::Select, so don't make any changes that block for long
# as new connections come we create a pair of entries in a hash %Con
# based on the hash of the filehandle, so $Con{$fh} has data for this
# connection. $Con{$fh}->{friend} is the partner socket for the smtp proxy.
# ->{ip} is the ip address of the connecting client
# ->{relayok} tells if we can relay mail for this client
# ->{getline} is a pointer to a function that should be called whan a
#               line of input is received for this filehandle
# ->{mailfrom} is the envelope sender (MAIL FROM: <address>)
# ->{outgoing} is a buffer for outgoing socket traffic (see $writable & &sendque)
# ->{rcpt} are the addresses from RCPT TO: <address> (space seperated)
# ->{header} is where the header (and eventually the first 10000 bytes) are stored
# ->{maillog} if present stream logging is enabled
# ->{maillogbuf} buffer for storing unwritten stream log while waiting for isspam decision
# ->{maillogfh} is the filehandle for logging lines to the maillog
# ->{mailloglength} is the length logged so far (we stop after 10000 bytes)

# After connection the {getline} field functions like a state machine
# redirecting input to subsequent handlers
#
#                                    whitebody -> getline
#                         getbody ->
#                                    error -> (disconnects)
# getline -> getheader ->
#                         whitebody -> getline
#
#                         error -> (disconnects)

# getline looks for MAIL FROM, RCPT TO, RSET
# getheader looks for a blank line then tests for whitelist / spamaddresses
# getbody looks for the . and calls isspam, the bayesian spam test
# whitebody wiats for . and redirects client to server
# error waits for . ignoring data from client (and finishes the maillog)

# the server has states like this:
#
# skipok -> reply
#
# skipok traps the 250 ok response from the NOOP Connection from
# reply echos server messages to the client
#      reply also looks for a 235 AUTH OK and sets {relayok}=1

use IO::Select;
use IO::Socket;

sub serviceCheck {}
sub d {}

eval(q[sub d {
 $time=gmtime(); $time=~s/... (...) (..) (........) ..(..)/$2 $1 $4 $3/;
 print DEBUG "$time <$_[0]>";
 }
]) if $DEBUG;

if($AsADaemon) {
 fork() && exit;
 close STDOUT;
 close STDERR;
 $silent=1;
}

if($pidfile) {open(F,">$base/$pidfile"); print F $$; close F;}

if($DEBUG) {open(DEBUG, ">$base/".time.".dbg"); binmode(DEBUG); my $oldfh = select(DEBUG); $| = 1; select($oldfh);}
if($logfile && open(LOG,">>$base/$logfile")) {my $oldfh = select(LOG); $| = 1; select($oldfh);}

if($AsAService) {
 eval(<<'EOT');
 use Win32::Daemon;
 mlog(0,'Starting as a Service');
 Win32::Daemon::StartService();

 # Wait until the service manager is ready for us to continue...
 while( SERVICE_START_PENDING != Win32::Daemon::State() ) { sleep( 1 ); }
 Win32::Daemon::State( SERVICE_RUNNING );

sub serviceCheck {
 d(50);
 if(Win32::Daemon::State() == SERVICE_STOP_PENDING ) {
  d(51);
  mlog(0,'Service Stopping');
  &saveWhitelist;
  Win32::Daemon::State( SERVICE_STOPPED );
  Win32::Daemon::StopService();
  exit;
 }
}

EOT
 print STDERR "error: $@\n" if $@;
 print LOG "error: $@\n" if $@;
}

init();
while(1) {
 &MainLoop;
}


sub init {
 mlog(0,"ASSP version $version initializing");

 # create folders if they're missing
 mkdir "$base/$spamlog",0777; mkdir "$base/$notspamlog",0777;
 my $dir=$correctedspam;
 $dir=~s/\/.*?$//;
 mkdir "$base/$dir",0777;
 mkdir "$base/$correctedspam",0777; mkdir "$base/$correctednotspam",0777;

 $SpamdbObject=tie %Spamdb,orderedtie,"$base/$spamdb" if $spamdb;
 mlog(0,"Warning: bayesian spam database is small or empty: '$base/$spamdb'") if $spamdb && -s "$base/$spamdb" < 10000;
 $DnsblObject=tie %Dnsbl,orderedtie,"$base/$dnsbl" if $dnsbl;
 mlog(0,"Warning: DNS blacklist database is small or empty: '$base/$dnsbl'") if $dnsbl && -s "$base/$dnsbl" < 10000;
 $WhitelistObject=tie %Whitelist,orderedtie,"$base/$whitelistdb";
 mlog(0,"Warning: Whitelist is small or empty: '$base/$whitelistdb' (ignore if this is a new install)") if $whitelistdb && -s "$base/$whitelistdb" < 1000;
 $RedlistObject=tie %Redlist,orderedtie,"$base/$redlistdb";
 $GoodhostsObject=tie %Goodhosts,orderedtie,"$base/goodhosts" unless $NoGoodhosts;

 $readable = new IO::Select( );
 $writable = new IO::Select( );

 $lsn = newListen($listenPort,\&NewSMTPConnection);
 $WebSocket = newListen($webAdminPort,\&NewWebConnection);
 mlog(0,"Listening for mail connections at $listenPort and admin connections at $webAdminPort");
 # handle the possible relayhost / smarthost option
 if($relayHost && $relayPort) {
  $Relay=newListen($relayPort,\&NewSMTPConnection);
  mlog(0,"Listening for relay connections at $relayPort");
 }

 $nextNoop=time;
 $endtime=$nextNoop+$RestartEvery;
 $saveWhite=$nextNoop+$UpdateWhitelist;

 $SIG{INT}=sub {mlog(0,'Sig INT'); &SaveWhitelist; kill 6,$$ if $ENV{windir}; exit;};
 $SIG{TERM}=sub {mlog(0,'Sig TERM'); &SaveWhitelist; exit;};

 my ($uid,$gid)=getUidGid($runAsUser,$runAsGroup) if ($runAsUser || $runAsGroup);
 if($ChangeRoot) {
  my $chroot;
  eval('$chroot=chroot($ChangeRoot)');
  if($@) {
   mlog('',"Request to change root to '$ChangeRoot' failed: $@");
   die "Request to change root to '$ChangeRoot' failed: $@";
  } elsif(! $chroot) {
   mlog('',"Request to change root to '$ChangeRoot' did not succeed: $!");
   die "Request to change root to '$ChangeRoot' did not succeed: $!";
  } else {
   mlog('',"Successfully changed root to '$ChangeRoot'");
  }
 }
 switchUsers($runAsUser,$runAsGroup) if ($runAsUser || $runAsGroup);

 $Stats{starttime}=time;
 $Stats{version}=$version;
 $Stats{pid}=$$;
 &ResetStats;

 mlog(0,"Starting");
}

sub newListen {
 my($port,$handler)=@_;
 my ($interface,$p)=$port=~/(.*):(.*)/;
 my $s;
 if($interface) {
  print "i=$interface p=$p ($port)\n";
  $s = new IO::Socket::INET(Listen => 10, LocalPort => $p, Reuse=>1, LocalAddr => $interface);
 } else {
  $s = new IO::Socket::INET(Listen => 10, LocalPort => $port, Reuse=>1);
 }
 if(! $s) {
  mlog('',"Couldn't create server socket on port '$port' -- maybe another service is running or I'm not root (uid=$>)?");
  #die "Couldn't create server socket on port '$port' -- maybe another service is running or I'm not root (uid=$>)?\n";
  return undef;
 }
 $SocketCalls{$s}=$handler;
 $readable->add($s); # add to select list
 $s;
}

sub getUidGid { my ($uname,$gname)=@_;
 return if $AsAService;
 eval('getgrnam(root);getpwnam(root);');
 if($@) {
  # windows pukes "unimplemented" for these -- just skip it
  mlog('',"Warning: uname and/or gname are set ($uname,$gname) but getgrnam / getpwnam give errors: $@");
  return;
 }
 my $gid;
 if($gname) {
  $gid = getgrnam($gname);
  if(defined $gid) {
  } else {
   mlog('',"Could not find gid for group '$gname' -- not switching effective gid -- quitting");
   die "Could not find gid for group '$gname' -- not switching effective gid -- quitting";
  }
 }
 my $uid;
 if($uname) {
  $uid = getpwnam($uname);
  if(defined $uid) {
  } else {
   mlog('',"Could not find uid for user '$uname' -- not switching effective uid -- quitting");
   die "Could not find uid for user '$uname' -- not switching effective uid -- quitting";
  }
 }
 ($gid,$uid);
}

sub switchUsers { my ($uid,$gid)=@_;
 return if $AsAService;
 $>=0;
 if($> != 0) {
  mlog('',"requested to switch to user/group '$uname/$gname' but cannot set effective uid to 0 -- quitting; uid is $>");
  die "requested to switch to user/group '$uname/$gname' but cannot set effective uid to 0 -- quitting; uid is $>\n";
 }
 $<=0;
 if($gid) {
  $)="$gid $gid";
  if($)+0==$gid) {
   mlog('',"Switched effective gid to $gid ($gname)");
  } else {
   mlog('',"Failed to switch effective gid to $gid ($gname) -- effective gid=$) -- quitting");
   die "Failed to switch effective gid to $gid ($gname) -- effective gid=$) -- quitting";
  }
  $(="$gid $gid";
  if($(+0==$gid) {
   mlog('',"Switched real gid to $gid ($gname)");
  } else {
   mlog('',"Failed to switch real gid to $gid ($gname) -- real uid=$(");
  }
 }
 if($uid) {
  # do it both ways so linux and bsd are happy
  $<=$uid; $>=$uid; $<=$uid; $>=$uid;
  if($>==$uid) {
  mlog('',"Switched effective uid to $uid ($uname)");
  } else {
   mlog('',"Failed to switch effective uid to $uid ($uname) -- real uid=$< -- quitting");
   die "Failed to switch effective uid to $uid ($uname) -- real uid=$< -- quitting";
  }
  if($<==$uid) {
   mlog('',"Switched real uid to $uid ($uname)");
  } else {
   mlog('',"Failed to switch real uid to $uid ($uname) -- real uid=$<");
  }
 }
}

sub MainLoop {
 my $wait=7; # keep it short enough for servicecheck to be called regularly
 my ($canread,$canwrite)=IO::Select->select($readable,$writable,undef,$wait);
 foreach $fh (@$canwrite) {
  my $l=length($Con{$fh}->{outgoing});
  print DEBUG "$fh $Con{$fh} l=$l\n" if $DEBUG;# '$Con{$fh}->{outgoing}'\n";
  if(length($Con{$fh}->{outgoing})) {
   my $written=syswrite($fh,$Con{$fh}->{outgoing},$OutgoingBufSize);
   if($DEBUG) {
   	if($written < 200) {
     print DEBUG "wrote: ($written)<",substr($Con{$fh}->{outgoing},$written),">\n";
    } else {
     print DEBUG "wrote: ($written)<toolong>\n";
    }
   }
   $Con{$fh}->{outgoing}=substr($Con{$fh}->{outgoing},$written);
   $l=length($Con{$fh}->{outgoing});
   # test for highwater mark
   if($written>0 && $l < $OutgoingBufSize && $Con{$fh}->{paused}) {
    $Con{$fh}->{paused}=0;
    $readable->add($Con{$fh}->{friend});
   }
  }
  if(length($Con{$fh}->{outgoing})==0) {
   $writable->remove($fh);
  }
 }
 foreach $fh (@$canread) {
  #print "fh=$fh\n";
  $SocketCalls{$fh}->($fh) if $fh && $SocketCalls{$fh};
 }
 d(6);

 serviceCheck(); # for win32 services

 # timer related issues
 $time=time;
 if($time - $ltime == $wait) {
  # noactive connections -- check for maintenance
  d(8);
  if($time >= $saveWhite) {
   d(9);
   &SaveWhitelist;
   $saveWhite=$time+$UpdateWhitelist;
  }
  if($RestartEvery && $time >= $endtime) {
   # time to quit -- after endtime and we're bored.
   $opencon=(keys %Con);
   if($opencon == 0) {
    &SaveWhitelist;
    mlog(0,"Restarting");
    exit 1;
   }
  }
  uploadStats() if($totalizeSpamStats && $time >= $Stats{nextUpload});
 }
 $ltime=$time;
}
d(11);
# Never reached...

sub SaveWhitelist {
 d(35);
 mlog(0,"Saving Whitelist");
 $WhitelistObject->flush();
 $RedlistObject->flush();
 $GoodhostsObject->flush() unless $NoGoodhosts;
}

sub mlog {
 my $fh=shift;
 d(34);
 my $m=localtime();
 $m=~s/^... (...) +(\d+) (\S+) ..(..)/$1-$2-$4 $3/;

 my $t=int(time()/(14*24*3600)); # roll log every 14 days
 if($logfile && $mlogLastT && $t != $mlogLastT ) {
  # roll the log
  my ($sec,$min,$hour,$mday,$mon,$year) = localtime(time);
  $mon++; $year-=100;
  my $mm=sprintf("%02d-%02d-%02d",$year,$mon,$mday);
  $logfile=~s/[0-9\-]*\.txt/$mm.txt/;
  $Config{logfile}=$logfile;
  print LOG "$m: Rolling log file -- new file is '$logfile'\n";
  print "$m: Rolling log file -- new file is '$logfile'\n" unless $silent;
  close LOG;
  if(open(LOG,">>$base/$logfile")) {my $oldfh = select(LOG); $| = 1; select($oldfh);}
  print LOG "$m: new log file\n";
  SaveConfig();
 }
 $mlogLastT=$t;
 if($fh && $Con{$fh}) {
  $m.= " $Con{$fh}->{ip} <$Con{$fh}->{mailfrom}>";
  my ($to) = $Con{$fh}->{rcpt}=~/(\S+)/;
  $m.= " to: $to" if $to;
 }
 $m.= " $_[0]\n";
 print $m unless $silent;
 print DEBUG $m if $DEBUG;
 print LOG $m if $logfile;
}

#####################################################################################
#                Socket handlers


sub NewSMTPConnection {
 my $fh=shift;
 my ($client, $server, $destination);
 if($fh==$lsn) {
  d(1);
  $client=$lsn->accept;
  $destination=$smtpDestination;
 } else { # must be a relay connection
  d(101);
  $client=$Relay->accept;
  $destination=$relayHost;
 }
 return unless $client;
 $server=new IO::Socket::INET(Proto=>'tcp',PeerAddr=>$destination,Timeout=>2);
 #print "friend=$server\n";
 #$NewCon=1;
 if(! $server) {
  mlog('',"Couldn't create server socket to $destination -- aborting connection");
  $client->close();
  next;
 }
 addfh($client,\&getline,$server);
 if($sendNoopInfo) {
  addfh($server,\&skipok,$client);
 } else {
  addfh($server,\&reply,$client);
 }
 #$ffh='aaaa' unless $ffh; open($ffh,">e:/spamdb/$ffh") &&  ($Con{$client}->{ffh}=$ffh) && binmode($ffh); $ffh++;
 my $ip=$Con{$client}->{ip}=$client->peerhost();
 print DEBUG "Connected: $client -- $server\n" if $DEBUG;
 $Stats{connects}++;
 if(ok2Relay($ip) || $fh == $Relay ) {
  $Con{$client}->{relayok}=1;
  #mlog($client,"relaing ok");
  print DEBUG "$client relaying ok: $ip\n" if $DEBUG;
 }
 $time=gmtime(); $time=~s/... (...) (..) (........) ..(..)/$2 $1 $4 $3/;
 $Con{$client}->{rcvd}="Received: from $ip ([$ip] helo=) by $myName ; $time -0000\r\n";
 print DEBUG  "\n* connect ip=$Con{$client}->{ip} relay=<$Con{$client}->{relayok}> *\n" if $DEBUG;
 mlog(0, "Connected: $ip:".$client->peerport());
 $Con{$server}->{noop}="NOOP Connection from: $ip, $time -0000 relayed by $myName\r\n" if $sendNoopInfo;
}

sub SMTPTraffic {
 my $fh=shift;
 my $buf;
 if($fh->sysread($buf,4096)>0) {
  d(2);
  $this=$Con{$fh};
  $buf=$this->{_}.$buf;
  if((my $sb=$this->{skipbytes})>0) {
   # support for XEXCH50
   # thankyou Microsoft for making my life miserable
   my $l=length($buf);
   print DEBUG "skipbytes=$sb l=$l -> " if $DEBUG;
   if($l >= $sb) {
    sendque($this->{friend},substr($buf,0,$sb)); # send the binary chunk on to the server
    $buf=substr($buf,$sb);
    delete $this->{skipbytes};
   } else {
    sendque($this->{friend},$buf); # send the binary chunk on to the server
    $this->{skipbytes}=$sb-=length($buf);
    $buf='';
   }
   print DEBUG "skipbytes=$this->{skipbytes}\n" if $DEBUG;
  }
  $bn=$lbn=-1;
  while(($bn=index($buf,"\n",$bn+1)) >= 0) {
   $s=substr($buf,$lbn+1,$bn-$lbn);
   print DEBUG  "doing <$s>\n" if $DEBUG;
   #print $fgfh $s if $fgfh=$Con{$fh}->{ffh};
   Maillog($fh,$s) if $Con{$fh}->{maillog};
   $Con{$fh}->{getline}->($fh,$s);
   last unless $Con{$fh};  # it's possible that the connection can be deleted while there's still something in the buffer
   $lbn=$bn;
  }
  d(3);
  if($Con{$fh}) {
   ($this->{_})=substr($buf,$lbn+1);
   if(length($this->{_}) > 10000) {
    d(4);
    Maillog($fh,$this->{_}) if $Con{$fh}->{maillog};
    $Con{$fh}->{getline}->($fh,$this->{_});
    ($this->{_})='';
   }
  }
 } else {
  d(5);
  done($fh);
 }
}

sub check4update {
 # only check every 15 seconds
 my $fil=shift;

 return if $check4updateTime{$fil} + 15 > time;
 $check4updateTime{$fil}=time;

 my @s=stat(${$fil});
 my $mtime=$s[9];
 if ($mtime != $FileUpdate{$fil}) {
  # reload
  $FileUpdate{$fil}=$mtime;
  open(F,"<${$fil}");
  local $/="\n";
  my $l; my %h;
  while($l=<F>) {
   chop($l);
   next unless $l;
   $h{lc $l}=1;
   print "$fil: <$l>=1\n";
  }
  close F;
  %{$fil}=%h;
 }
}

sub ok2Relay {
 my $ip=shift;
 return 1 if $acceptAllMail && $ip=~/^($acceptAllMail)/i;

 if($relayHostFile) {
  check4update(relayHostFile);
  return 1 if $relayHostFile{$ip};
 }

 # If you want to allow pop before smtp uncomment this line and implement the PopB4SMTP sub below
 #return 1 if PopB4SMTP($ip);

 # failed all tests -- return 0
 0;
}

# if you want to allow pop before smtp authentication then umcomment the line above
# where this sub is called and edit this sub to be appropriate for your implementation
sub PopB4SMTP {
 #my $ip=shift;
 #use DB_File ;
 #my %hash;
 #tie %hash,  'DB_File', [$filename, $flags, $mode, $DB_HASH] ;
 #if($hash{$ip} > 0) {
 # return 1;
 #} else {
 # return 0;
 #}
}

sub NewWebConnection {
 my $s=$WebSocket->accept;
 return unless $s;
 if($allowAdminConnectionsFrom) {
  my $ip=$s->peerhost();
  mlog('',"admin connection from $ip");
  if($ip!~/^($allowAdminConnectionsFrom)/i) {
   mlog('',"admin connection from $ip rejected by allowAdminConnectionsFrom");
   $s->close();
   return;
  }
 }
 $readable->add($s);
 $SocketCalls{$s}=\&WebTraffic;
}

sub WebTraffic {
 my $fh=shift;
 if($fh->sysread($buf,40960)>0) {
  local $_=$WebCon{$fh}.=$buf;
  if(length($_) > 1030000) {
   # throw away connections longer than 1M to prevent flooding
   WebDone($fh);
   return;
  }
  if(/Content-length: (\d+)/i) {
   webRequest($fh,$1,$2) if	/(.*\n)\r?\n\r?(.*)/s;
  } elsif(/\n\r?\n/) {
   webRequest($fh,$_);
  }
 } else {
  # connection closed
  WebDone($fh);
 }
}

sub WebDone {
 my $fh=shift;
 delete $SocketCalls{$fh};
 delete $WebCon{$fh};
 $readable->remove($fh);
}

# done with a file handle -- close him and his friend
sub done {
 my $fh=shift;
 d(12);
 done2($Con{$fh}->{friend});
 done2($fh);
}

# close a file handle & clean up associated records
sub done2 {
 my $fh=shift;
 d(13);
 return unless $fh;
 my $this=$Con{$fh};
 return unless $this;
 print DEBUG  "closing $fh\n" if $DEBUG;
 # close the maillog if it's still open
 my $f=$this->{maillogfh};
 close $f if $f;
 #$f=$this->{ffh}; close $f if $f;
 # remove from the select structure
 delete $SocketCalls{$fh};
 $readable->remove($fh);
 $writable->remove($fh);
 # close it
 $fh->close;
 # delete the Connection data
 delete $Con{$fh};
}

# adding a socket to the Select structure and Con hash
sub addfh {
 my ($fh,$getline,$friend) =@_;
 d(14);
 $SocketCalls{$fh}=\&SMTPTraffic;
 $readable->add($fh);
 binmode($fh);
 $Con{$fh}={};
 my $this=$Con{$fh};
 $this->{getline}=$getline;
 $this->{friend}=$friend;
 #print "add $fh: f=$friend\n";
}

# sendque enques a string for a socket
sub sendque {
 my ($fh,$message)=@_;
 my $l=length($message);
 print DEBUG "sq: $fh l=$l\n" if $DEBUG;#'$message'\n";
 #print "sq: $message\n";
 return unless $fh && $Con{$fh};
 $writable->add($fh);
 $Con{$fh}->{outgoing}.=$message;
 if(!$Con{$fh}->{paused} && length($Con{$fh}->{outgoing}) > $OutgoingBufSize) {
  $Con{$fh}->{paused}=1;
  print DEBUG "pausing\n" if $DEBUG;
  $readable->remove($Con{$fh}->{friend});
 }
}

#####################################################################################
#                SMTP stuff

sub localmail {
 my $h=shift;
 print "h=$h ld=$localDomains ldf=$localDomainsFile\n";
 return 1 if $localDomains && $h=~/^($localDomains)$/i;
 if($localDomainsFile) {
  check4update(localDomainsFile);
  return 1 if $localDomainsFile{lc $h};
 }
 0;
}

# a line of input has been received from the smtp client
sub getline {
 my($fh,$l)=@_;
 d(15);
 my $this=$Con{$fh};
 my $server=$this->{friend};
 print DEBUG  "gl: <$l>\n" if $DEBUG;
 if($l=~/^ *(helo|ehlo) .*?([^<>,;"'\(\)\s]+)/i) {
  my $helo = $2;
  my $helo2 = $helo;
  $helo=~s/(\W)/\\\1/g;
  $l =~ s/$helo/$myName/ ;
  $this->{helo}=$helo2;
  $this->{rcvd}=~s/=/=$helo2/;
 } elsif($l=~/mail from:.*?([^<>,;"'\(\)\s]+\@[^<>,;"'\(\)\s\?]+)/i) {
  $this->{rcpt}=$this->{header}=''; # reset everything else
  $this->{mailfrom}=$1;
  $this->{noprocessing}=0;
 } elsif($l=~/rcpt to:.*?([^<>,;"'\(\)\s]+\@)([^<>,;"'\(\)\s\?]+)/i) {
  my($u,$h)=($1,$2);
  #print "rcp <$u$h> r=<$this->{'relayok'}> $fh\n";
  if(!($this->{'relayok'}) && (! localmail($h) || $u=~/\%/)) {
   sendque($fh, $NoRelaying."\r\n");
   mlog($fh,"relay attempt blocked for: $u$h");
   $Stats{norelays}++;
   return;
  }
  if($noProcessing && $h=~/^($localDomains)$/i) {
   if($u=~/^($noProcessing)\@/i) {
    $this->{noprocessing}|=1;
   } else {
    $this->{noprocessing}|=2;
   }
  }
  if($EmailInterfaceOk && $this->{relayok} && $h=~/^($localDomains)$/i) {
   if(lc $u eq lc "$EmailSpam\@") {
    $this->{getline}=\&SpamReport;
    mlog($fh,"Email spamreport");
    sendque($fh,"250 ok\r\n");
    return;
   } elsif(lc $u eq lc "$EmailHam\@") {
    $this->{getline}=\&HamReport;
    mlog($fh,"Email hamreport");
    sendque($fh,"250 ok\r\n");
    return;
   } elsif(lc $u eq lc "$EmailWhitelist\@") {
    $this->{getline}=\&EmailWhite;
    mlog($fh,"Email Whitelist");
    for $a (split(/ /,$this->{rcpt})) {EmailWhiteAdd($a,$this)};
    sendque($fh,"250 ok\r\n");
    return;
   }
  }
  $this->{rcpt}.="$u$h ";
 } elsif( $l=~/^ *XEXCH50 +(\d+)/i ) {
  $this->{skipbytes}=$1;
  print DEBUG "XEXCH50 b=$1\n" if $DEBUG;
 } elsif( $l=~/^ *DATA/i ) {
  if($this->{noprocessing}==1) {
   # all addresses are on no-processing list
   $this->{getline}=\&whitebody;
   mlog($fh,"message proxied without processing");
  } else {
   MaillogStart($fh); # notify the stream logging to start logging
   $this->{getline}=\&getheader;
  }
  $Stats{messages}++;
 } elsif( $l=~/^ *RSET/i ) {
  $this->{mailfrom}=$this->{rcpt}=$this->{header}='';
  $this->{noprocessing}=0;
 }
 sendque($server, $l);
}

# get the header part of the DATA
sub getheader {
 my($fh,$l)=@_;
 d(18);
 my $this=$Con{$fh};
 $this->{header}.=$l;
 if($l=~/^[\r\n]*$/) {
  # header is done
  #print "header is done\n";
  if(onwhitelist($fh,$this->{header})) {
   Maillog($fh,'',2); # tell maillog this isn't spam
   $SpamProb=0; addSpamProb($fh);
   if($BlockAllExes) {
    $this->{getline}=\&whitebodyNoExe;
   } else {
    isnotspam($fh);
   }
   mlog($fh,'local or whitelisted');
   $Goodhosts{$this->{ip}}=time if(!$NoGoodhosts && ! $this->{relayok});
  } elsif( $blackListedDomains && ($this->{mailfrom}=~/($blackListedDomains)$/i || $this->{senders}=~/($blackListedDomains) /i)) {
   Maillog($fh,'',3); # tell maillog this is spam
   seterror($fh,$SpamError);
   $Stats{spams}++;
   mlog($fh,'has blacklisted domain');
  } elsif(hasspamaddress($this->{header},$this->{rcpt})) {
   Maillog($fh,'',3); # tell maillog this is spam
   seterror($fh,$SpamError);
   $Stats{spams}++;
   mlog($fh,'has spam address');
  #} elsif($spamLovers && allLoveSpam($this->{rcpt})) {
  # Maillog($fh,'',4); # tell maillog to forget the message
  # isnotspam($fh);
  # mlog($fh,"rcpt to only spamlovers: $this->{rcpt}");
  } else {
   # not whitelisted, not spamaddress, let's see if it's spam or not
   $this->{getline}=\&getbody;
  }
 }
}

sub allLoveSpam {
 my $rcpt=shift;
 my $c=0;
 for (split(' ',$rcpt)) {
  return 0 unless /^($spamLovers)\@/i;
  $c++;
 }
 $c;
}

# the message is not spam -- route it to the server
sub isnotspam {
 my $fh=shift;
 d(23);
 my $done=shift;
 my $this=$Con{$fh};
 my $server=$this->{friend};

 my $m=$this->{header};
 sendque($server, $m);
 $this->{header}='';
 if($done) {
  $this->{getline}=\&getline;
 } else {
  $this->{getline}=\&whitebody;
 }
}

# the message is non spam -- just relay it to the server
sub whitebody { my($fh,$l)=@_;
 my $this=$Con{$fh};
 d(25);
 my $server=$this->{friend};
 if( $l=~/^\.[\r\n]*$/) {
  $this->{getline}=\&getline;
 }
 sendque($server, $l);
}

# the message is non spam -- check if it is executable
sub whitebodyNoExe { my($fh,$l)=@_;
 my $this=$Con{$fh};
 d(255);
 my $b=$this->{header}.=$l;
 my $done=$l=~/^\.[\r\n]*$/;
 if( $done || length($b) >=10000) {
  if($badattach && $b=~/$badattach/i) {
   Maillog($fh,'',4); # tell maillog to ignore the message
   seterror($fh,$AttachmentError,$done);
   $Stats{viri}++;
   mlog($fh,"bad attachment (local/white)");
  } else {
   isnotspam($fh,$done);
  }
 }
}

# the message may or may not be spam -- get the body and test it
sub getbody { my($fh,$l)=@_;
 d(27);
 my $this=$Con{$fh};
 my $b=$this->{header}.=$l;
 my $done=$l=~/^\.[\r\n]*$/;
 if( $done || length($b) >=10000) {
  if($BlockExes && $badattach && $b=~/$badattach/i) {
   Maillog($fh,'',4); # tell maillog to ignore the message
   seterror($fh,$AttachmentError,$done);
   $Stats{viri}++;
   mlog($fh,"bad attachment");
  } elsif(isspam($b,$this->{ip})) {
   Maillog($fh,'',3); # tell maillog it is spam
   mlog($fh,"bayesian spam");
   if(!$NoGoodhosts && ! $Goodhosts{$this->{ip}}) {$Goodhosts{'spam'}++}
   $Stats{bspams}++;
   my $slok;
   if($TestMode || ($slok=$spamLovers && allLoveSpam($this->{rcpt})) ) {
    if($slok) {
     mlog($fh,"rcpt to only spamlovers: $this->{rcpt}");
     $Stats{spamlover}++;
    }
    # just testing -- pretend it's not spam
    addSpamProb($fh);
    $this->{header}=~s/\nSubject:/\nSubject: $spamSubject/i if $spamSubject;
    $this->{header}=~s/\nSubject:/\nX-Assp-Spam: YES\r\nSubject:/i if $AddSpamHeadder;
    isnotspam($fh,$done);
   } else {
    seterror($fh,$SpamError,$done);
   }
  } else {
   # skip saving bayesian nonspam -- prevents false negs from corrupting corpus
   Maillog($fh,'',4); # tell maillog to ignore the message
   addSpamProb($fh);
   if(!$NoGoodhosts && ! $Goodhosts{$this->{ip}}) {$Goodhosts{'notspam'}++}
   isnotspam($fh,$done);
   $Stats{bhams}++;
   mlog($fh,"message ok");
  }
 }
}

# reject the email
sub seterror {
 my($fh,$e,$done)=@_;
 d(28);
 my $this=$Con{$fh};
 $this->{error}=$e;
 if($done) {
  error($fh,".\r\n");
 } else {
  $this->{getline}=\&error;
  $this->{header}=''; # free up some memory
 }
 # detatch the friend -- closing connection to server & disgarding message
 done2($this->{friend});
 $this->{friend}=undef;
}

# ignore what's sent & give reason at the end.
sub error { my($fh,$l)=@_;
 d(29);
 my $this=$Con{$fh};
 if( $l=~/^\.[\r\n]*$/) {
  print $fh $this->{error}."\r\n";
  done2($fh);
 }
}

# filter off the 250 OK noop response and go to reply
sub skipok {
 d(291);
 my ($fh,$l)=@_;
 if($l=~/^250/) {
  $Con{$fh}->{getline}=\&reply;
 } else {
  reply(@_);
 }
}

# messages from the server get relayed to the client
sub reply {
 my($fh,$l)=@_;
 d(30);
 my $this=$Con{$fh};
 return unless $this;
 my $cli=$this->{friend};
 return unless $cli;

 if($l=~/^220/) {
  #print DEBUG "sending $this->{noop}" if $DEBUG && $this->{noop};
  sendque($fh,$this->{noop}) if $this->{noop};
  delete $this->{noop};
 } elsif($l=~/^235/) {
  # check for authentication response
  ${$Con{$cli}}{relayok}=1;
  print DEBUG  "$this->{ip}: authenticated\n" if $DEBUG;
  mlog($fh,"Authenticated");
 } elsif($l=~/^354/) {
  d(301);
 } elsif($l=~/^50[0-9]/) {
  if($this->{serverErrors}++ > $MaxErrors) {
   mlog($fh,"Max errors ($MaxErrors) exceeded -- dropping connection");
   done($fh);
  }
 }

 #print "r: $l ($fh) ($cli)\n";
 sendque($cli, $l);
}


#####################################################################################
#                Email Interface

# this mail isn't really a mail -- it's a spam report
sub SpamReport {
 my($fh,$l)=@_;
 my $this=$Con{$fh};
 if( $l=~/^ *DATA/i ) {
  $this->{getline}=\&SpamReport2;
  sendque($fh,"354 ok\r\n");
  return;
 } elsif( $l=~/^ *RSET/i ) {
  $this->{getline}=\&getline;
  sendque($this->{friend},"RSET\r\n");
 } elsif( $l=~/^ *XEXCH50 +(\d+)/i ) {
  $this->{skipbytes}=$1;
  print DEBUG "XEXCH50 b=$1\n" if $DEBUG;
  sendque($fh,"354 Send binary data\r\n");
  return;
 }
 sendque($fh,"250 ok\r\n");
}

# we're getting the body of a spam report
sub SpamReport2 {
 my ($fh, $l)=@_;
 $Con{$fh}->{header}.=$l if length($Con{$fh}->{header}) < 10000;
 if($l=~/^\.[\r\n]/) {
  # we're done -- write the file & clean up
  my $this=$Con{$fh};
  my $sub=ReportMail($this->{header},$correctedspam);
  $this->{header}='';
  ReturnMail($this->{mailfrom},"Spam report: $sub",<<EOT);
You reported spam.
It now has been recorded.
Thank you very very much.

-----

These words not asked for.
Why are you in my email?
Blue can, fat pink meat.

bamboo flute on pond
angry man shakes fists at sky
e-mail clogged with spam

www.spamhaiku.com
EOT

  $this->{getline}=\&getline;
  sendque($this->{friend},"RSET\r\n");
 }
}

# this mail isn't really a mail -- it's a false-positive report
sub HamReport {
 my($fh,$l)=@_;
 my $this=$Con{$fh};
 if( $l=~/^ *DATA/i ) {
  $this->{getline}=\&HamReport2;
  sendque($fh,"354 ok\r\n");
  return;
 } elsif( $l=~/^ *RSET/i ) {
  $this->{getline}=\&getline;
  sendque($this->{friend},"RSET\r\n");
 } elsif( $l=~/^ *XEXCH50 +(\d+)/i ) {
  $this->{skipbytes}=$1;
  print DEBUG "XEXCH50 b=$1\n" if $DEBUG;
  sendque($fh,"354 Send binary data\r\n");
  return;
 }
 sendque($fh,"250 ok\r\n");
}

# we're getting the body of a ham report
sub HamReport2 {
 my ($fh, $l)=@_;
 $Con{$fh}->{header}.=$l if length($Con{$fh}->{header}) < 10000;
 if($l=~/^\.[\r\n]/) {
  # we're done -- write the file & clean up
  my $this=$Con{$fh};
  my $sub=ReportMail($this->{header},$correctednotspam);
  $this->{header}='';
  ReturnMail($this->{mailfrom},"Not-Spam report: $sub",<<EOT);
I made a mistake.
Blocking good mail is shameful.
Now I regain my honor.

-----

i face Spam's mirror.
Pink Buddha enlightens me
for now I see mapS

Shapely cube of SPAM
That hides universes in
A shiny tin can

www.spamhaiku.com
EOT

  $this->{getline}=\&getline;
  sendque($this->{friend},"RSET\r\n");
 }
}

sub ReportMail {
 my ($bod,$path)=@_;
 my ($sub)=$bod=~/Subject: (.*)/i;
 $bod=~s/^.*?\n\r?\n\s*//s;
 # remove the spam subject header addition if present
 my $spamsub=$spamSubject;
 if($spamsub) {
  $spamsub=~s/(\W)/\\$1/g;
  $bod=~s/Subject: $spamsub */Subject: /gi;
 }
 $bod=~s/X-Assp-Spam-Prob: .*\n//gi;
 if($bod=~/\nReceived: /) {
  $bod=~s/^.*?\nReceived: /Received: /s;
 } else {
  $bod=~s/^.*?\n((\w[^\n]*\n)*Subject:)/$1/si;
  $bod=~s/\n> /\n/g;
 }
 my $f;
 do {
  $f=int(rand()*100000000);
 } while(-e "$base/$path/$f.rpt");
 open(F,">$base/$path/$f.rpt"); binmode F;
 print F $bod;
 close F;
 $sub;
}

# we're receiving an email to add addresses to the whitelist
sub EmailWhite {
 my($fh,$l)=@_;
 d('EmailWhite');
 my $this=$Con{$fh};
 if( $l=~/^ *DATA/i ) {
  $this->{getline}=\&EmailWhiteBody;
  sendque($fh,"354 OK Send whitelist body\r\n");
  return;
 } elsif( $l=~/^ *RSET/i ) {
  $this->{getline}=\&getline;
  sendque($this->{friend},"RSET\r\n");
  return;
 } elsif( $l=~/^ *XEXCH50 +(\d+)/i ) {
  $this->{skipbytes}=$1;
  print DEBUG "XEXCH50 b=$1\n" if $DEBUG;
  sendque($fh,"354 Send binary data\r\n");
  return;
 } else {
  for $a ($l=~/([^<>,;"'\(\)\s]+\@[^<>,;"'\(\)\s\?]+)/g) {
   EmailWhiteAdd($a,$this);
   $this->{rcpt}.="$a ";
  }
 }
 sendque($fh,"250 OK\r\n");
}

sub EmailWhiteBody {
 my($fh,$l)=@_;
 #print "ew2($l) ";
 d('EmailWhiteBody');
 my $this=$Con{$fh};
 if($l=~/^\.[\r\n]/) {
  # mail summary report
  ReturnMail($this->{mailfrom},"Whitelist Addition Report",<<EOT);
Note -- Mail not delivered to the following recipients:
$this->{rcpt}

$this->{report}
EOT
  delete $this->{report};
  $this->{getline}=\&getline;
  sendque($this->{friend},"RSET\r\n");
 } elsif($l=~/message-id:/i) {
  # ignore
 } else {
  for $a ($l=~/([^<>,;"'\(\)\s:]+\@[^<>,;"'\(\)\s\?]+)/g) {
   EmailWhiteAdd($a,$this);
  }
 }
}

sub EmailWhiteAdd {
 my ($a,$this)=@_;
 my $t=time;
 if($a=~/\@($localDomains)$/) {
  $this->{report}.="$a: Cannot add local users to whitelist\n";
 } elsif( $Whitelist{lc $a} ) {
  $this->{report}.="$a: already on whitelist\n";
  mlog(0,"email whitelist addition: $a");
  $Whitelist{lc $a}=$t;
 } else {
  $this->{report}.="$a: added\n";
  mlog(0,"email new whitelist addition: $a");
  $Whitelist{lc $a}=$t;
 }
}


sub ReturnMail {
 my $s=new IO::Socket::INET(Proto=>'tcp',PeerAddr=>$smtpDestination,Timeout=>2);
 if(! $s) {
  mlog('',"Couldn't create server socket to $smtpDestination -- aborting ReturnMail connection");
  return;
 }
 addfh($s,\&RMhelo);
 my $this=$Con{$s};
 $this->{to}=shift;
 $this->{from}=$EmailFrom;
 $this->{subject}=shift;
 $this->{body}=shift;
}

sub RMhelo { my ($fh,$l)=@_;
 if($l=~/^ *5/) {
  RMabort($fh,"helo Expected 220, got: $l");
 } elsif($l=~/^ *220 /) {
  sendque($fh,"HELO $myName\r\n");
  $Con{$fh}->{getline}=\&RMfrom;
 }
}
sub RMfrom { my ($fh,$l)=@_;
 if($l=~/^ *5/) {
  RMabort($fh,"from Expected 250, got: $l");
 } elsif($l=~/^ *250 /) {
  sendque($fh,"MAIL FROM: $Con{$fh}->{from}\r\n");
  $Con{$fh}->{getline}=\&RMrcpt;
 }
}
sub RMrcpt { my ($fh,$l)=@_;
 if($l=~/^ *5/) {
  RMabort($fh,"rcpt Expected 250, got: $l");
 } elsif($l=~/^ *250 /) {
  sendque($fh,"RCPT TO: <$Con{$fh}->{to}>\r\n");
  $Con{$fh}->{getline}=\&RMdata;
 }
}
sub RMdata { my ($fh,$l)=@_;
 if($l=~/^ *5/) {
  RMabort($fh,"data Expected 250, got: $l");
 } elsif($l=~/^ *250 /) {
  sendque($fh,"DATA\r\n");
  $Con{$fh}->{getline}=\&RMdata2;
 }
}
sub RMdata2 { my ($fh,$l)=@_;
 if($l=~/^ *5/) {
  RMabort($fh,"data2 Expected 354, got: $l");
 } elsif($l=~/^ *354 /) {
  my $date=gmtime();
  $date=~s/(\w+) +(\w+) +(\d+) +(\S+) +(\d+)/$1, $3 $2 $5 $4/;
  my $this=$Con{$fh};
  sendque($fh,<<EOT);
From: $this->{from}\r
To: $this->{to}\r
Subject: $this->{subject}\r
Date: $date +0000\r
\r
$this->{body}\r
.\r
EOT
  $Con{$fh}->{getline}=\&RMdone;
 }
}
sub RMdone { my ($fh,$l)=@_;
 if($l=~/^ *5/) {
  RMabort($fh,"done Expected 250, got: $l");
 } elsif($l=~/^ *250 /) {
  done2($fh); # close and delete
 }
}
sub RMabort {mlog(0,"RMabort: $_[1]"); done2($_[0]);}


#####################################################################################
#                SPAM Detection

# check if the message is spam, based on bayesian factors in $Spamdb
sub isspam {
 my $msg=$_[0];
 return $SpamProb=0 if $whiteRe && $msg=~/$whiteRe/is;
 return $SpamProb=1 if $blackRe && $msg=~/$blackRe/is;
 local $b=clean($_[0]);
 return $SpamProb=0 if $whiteRe && $b=~/$whiteRe/is;
 return $SpamProb=1 if $WhiteOnly || $blackRe && $b=~/$blackRe/is;
 my $ip=$_[1];
 d(31);
 my $ip3=$ip;
 $ip3=~s/(\d+\.\d+\.\d+).*/$1/;
 my ($v,$lt,$llt,$t,%seen);
 my @t;
 #print "ip=$ip ($Dnsbl{$ip}) ip3=$ip3 ($Dnsbl{$ip3})\n";
 if(defined($Dnsbl{$ip}) || defined($Dnsbl{$ip3})) {
  mlog('',"$ip dnsbl hit");
  push(@t,0.97);
  push(@t,0.97);
 }
 # the theory here is this: Most good mail is from hosts that send whitelisted mail
 # For hosts that haven't sent us good mail we'll track the ratio of spam vs notspam.
 # our experimental results show it's about 90% spam, but your results will be different
 if(!$NoGoodhosts && ! $Goodhosts{$ip}) {
  my $p=$Goodhosts{'spam'}/($Goodhosts{'notspam'}+$Goodhosts{'spam'}+1);
  #mlog(0,"p=$p $Goodhosts{'notspam'}+$Goodhosts{'spam'}");
  if($p > 0.65) {
   push(@t,$p);
   push(@t,$p);
  }
 }
 while($b=~/([-\$A-Za-z0-9\'\.\,!]+)/g) {
  next if length($1) > 20;
  $llt=$lt; $lt=$t; $t=lc $1;
  #next if $t=~/^\d/; # ignore numbers
  $t=~s/[,.']+$//; $t=~s/!!!+/!!/g; $t=~s/--+/-/g;
  #push(@t,$v) if ($v=$Spamdb{$t});
  #my $j="$lt $t"; #print "$t:$spa{$t} ";
  #push(@t,$v) if ($v=$Spamdb{$j});
  my $j="$llt $lt $t"; #print "$t:$spa{$t} ";
  next if $seen{$j}++ >1; # first two occurances are significant
  push(@t,$v) if ($v=$Spamdb{$j});
 }
 @t=sort {abs($b-.5)<=>abs($a-.5)} @t;
 @t=@t[0..30];
 #print "\n",join(',',@t),"\n";
 my $p1=1; my $p2=1; for $p (@t) {if($p) {$p1*=$p; $p2*=(1-$p);}}
 $p1=$p1/($p1+$p2);
 #printf("isspam= %.4f\n",$p1);
 $SpamProb=$p1;
 $p1<.6? 0: 1;
}

# attach a header line to the message if the config option is set
sub addSpamProb {
 return unless $AddSpamProbHeadder;
 my $fh=shift;
 return if $NoExternalSpamProb && $Con{$fh}->{relayok};  ## rainer likes it this way
 my $spamprob=sprintf("X-Assp-Spam-Prob: %.5f\r\n",$SpamProb);
 $Con{$fh}->{header}=~s/^(.*?)(\r?\n)\r?\n/$1$2$spamprob$2/s;
}

# see if the address in the mailfrom is on the whitelist
# meanwhile update the whitelist if that seems appropriate
sub onwhitelist {
 my($fh,$header,$a)=@_;
 d(32);
 my $this=$Con{$fh};

 $a=lc $this->{mailfrom};
 return 0 if !$a || $Redlist{$a} || ($redRe && $header=~/$redRe/is);
 my $whitelisted=$this->{relayok};
 $Stats{locals}++ if $whitelisted;
 return $whitelisted unless defined $a; # don't add to the whitelist unless there's a valid envelope -- prevent bounced mail from adding to the whitelist
 my %senders;
 unless($whitelisted) {
  $senders{$a}=1;
  while($header=~/\n(from|sender|reply-to|errors-to|list-\w+):.*?([^:<>,;"'\(\)\s]+\@[^<>,;"'\(\)\s\?]+\.[^<>,;"'\(\)\s\?]+)/ig) {
   #print "$1:$2  ";
   $senders{lc $2}=1;
  }
  for $a (keys %senders) {
   #print "wl: $a '$Whitelist{$a}'  ";
   return 0 if $a && $Redlist{$a};
   next if $a=~/\@($localDomains)$/i || $a eq '';
   if($Whitelist{$a} || $whiteListedDomains && $a=~/($whiteListedDomains)$/i) {
    $whitelisted=1;
    #print "**yes whitelist**\n";
    last;
   }
  }
  $this->{senders}=join(' ',keys %senders)." ";
  $Stats{whites}++ if $whitelisted;
 }
 if($whitelisted) {
  # keep the whitelist up-to-date
  my %a=%senders;
  my $t=time;
  $a{$a}=1;
  $header=~s/\n\s+/ /g;
  while($header=~/\n(to|cc): (.*)/ig) {
   #print "h=$2\n";
   my $a=$2;
   while($a=~/([^<>,;"'\(\)\s]+\@[^<>,;"'\(\)\s\?]+)/g) {
    $a{lc $1}=1;
   }
  }
  #print "r=$this->{rcpt}\n";
  for $a (split(' ',lc $this->{rcpt})) {
   $a{$a}=1;
  }
  for $a (keys %a) {
   #print "wl+$a ";
   next if $a=~/\@($localDomains)$/i || ! $a;
   #print "notlocal ";
   next if $whiteListedDomains && $a=~/($whiteListedDomains)$/i;
   #print "whitelist+ $a\n";
   mlog($fh,"whitelist addition: $a") unless $Whitelist{$a};
   $Whitelist{$a}=$t;
  }
  #print "you are on the whitelist or are a relay client\n";
  return 1;
 }
 #print "you are NOT on the whitelist\n";
 0;
}

# clean up source email
sub clean {
 local $_="\n".shift;
 # replace &#ddd encoding
 s/&#(\d{1,3});?/chr($1)/ge;
 #s/base64.{0,99}\n\n([a-zA-Z0-9+\/\n=]+)/base64decode($1)/gse;
 # replace base64 encoding
 s/\n([a-zA-Z0-9+\/=]{40,}\r?\n[a-zA-Z0-9+\/=\r\n]+)/base64decode($1)/gse;
 # clean up quoted-printable references
 if(/quoted-printable/) {
  s/=\r?\n//g;
  s/=([0-9a-fA-F]{2})/pack("C",hex($1))/gei;
 }
 s/(http:\/\/\S+)/fixurl($1)/ige;
 # strip out mime continuation
 s/.*---=_NextPart_.*\n//g;
 # mark the subject
 s/\nsubject: (.*)/fixsub($1)/ige;
 # remove received lines
 s/\n(received|Content-Type): .*(\n[\t ].*)*//ig;
 # remove other header lines
 s/\n[a-zA-Z\-]{2,40}: .*//g;
 # clean up &nbsp; and &amp;
 s/&nbsp;?/ /gi; s/&amp;?/and/gi;
 # clean up html stuff
 while(s/(\w+)(<[^>]*>)((<[^>]*>)*\w+)/$2$1$3/g){} # move html out of words
 s/<([biu]|strong)>/ boldifytext /gi; s/<\/?(p|br|div|t[dr])[^>]*>/\n/gi; s/<\/([biua]|font|strong)>//gi;
 s/<\/?(html|meta|head|body|span|o)[^>]*>//ig;
 # treat titles like subjects
 s/<title[^>]*>(.*?)<\/title>/fixsub($1)/ige;
 # remove style sheets
 s/<style[^>]*>.*?<\/style>//igs;
 # remove html comments
 s/<!.*?-->//gs; s/<![^>]*>//g;
 # look for random words
 s/ [bcdfghjklmnpqrstvwxz]{3}[bcdfghjklmnpqrstvwxz0-9]{3,}/ randword randword randword /gi;
 # look for linked images
 s/(<a[^>]*>[^<]*<img)/ linkedimage linkedimage linkedimage $1/gi;
 #open(F,">t"); print F $_; close F;
 $_;
}

sub fixurl {
 my $a=shift;
 $a=~s/%([0-9a-fA-F][0-9a-fA-F])/pack('C',hex($1))/ge;
 $a;
}

sub fixsub {
 my $s=shift;
 #print "$s=>";
 $s=~s/ {3,}/ lotsaspaces /g;
 $s=~s/(\S+)/ssub $1/g;
 #print "$s\n";
 "\nssub $s ssub";
}

sub base64decode {
 my $str = shift;
 my $res = "\n\n";
 $str =~ tr|A-Za-z0-9+/||cd;
 $str =~ tr|A-Za-z0-9+/| -_|;
 while ($str =~ /(.{1,60})/gs) {
  my $len = chr(32 + length($1)*3/4);
 $res .= unpack("u", $len . $1 );
 }
 $res;
}

sub hasspamaddress { my($header,$rcpt) =@_;
 return 0 unless $spamaddresses;
 print DEBUG  "has spam addr\n" if $DEBUG && " $rcpt\n$header"=~/[^\w\-\.]($spamaddresses)\@($localDomains)[^\w\-]/i;
 " $rcpt\n$header"=~/[^\w\-\.]($spamaddresses)\@($localDomains)[^\w\-]/i;
}

sub uploadStats {
 mlog(0,"Uploading stats");
 my $s=new IO::Socket::INET(Proto=>'tcp',PeerAddr=>"assp.sourceforge.net:80",Timeout=>2);
 if($s) {
  $Stats{timenow}=time;
  $Stats{testmode}=$TestMode;
  my $content=join("\001",%Stats);
  my $len=length($content);
  print $s "POST /cgi-bin/upload.pl HTTP/1.1
Host: assp.sourceforge.net
Content-Type: application/x-www-form-urlencoded
Content-Length: $len

$content";
  $s->close;
 }
 #resetStats();
 $Stats{nextUpload}=time+3600*8;
}

sub ResetStats {
 $Stats{nextUpload}=time+3600*8;
 $Stats{locals}=0;
 $Stats{whites}=0;
 $Stats{messages}=0;
 $Stats{hams}=0;
 $Stats{spams}=0;
 $Stats{bhams}=0;
 $Stats{bspams}=0;
 $Stats{viri}=0;
}

#####################################################################################
#                Maillog functions


# find an appropriate name for a maillog file
sub maillogFilename {my $isspam=shift;
 my @dirs=($notspamlog,$spamlog,$incomingOkMail);
 my $maillog=$dirs[$isspam];
 d(19);
 if($UseSubjectsAsMaillogNames) {
  my($sub)=$_[0]=~/Subject: (.*)/;
  $sub=~y/a-zA-Z0-9/_/cs;
  $sub.="--".(++$Counter);
  return "$base/$maillog/$sub";
 } else {
  my $fn=(time + $Counter++ ) % $MaxFiles;
  #print "isspam=$isspam; log to $maillog/$fn\n";
  "$base/$maillog/$fn";
 }
}


# integrated mail collection subroutine
sub MaillogStart {
 d(361);
 $Con{$_[0]}->{maillog}=1 unless $NoMaillog;
 $Con{$_[0]}->{maillogbuf}=$Con{$_[0]}->{header}=$Con{$_[0]}->{rcvd};
}
sub Maillog {
 my ($fh,$text,$parm)=@_;
 # parm = 2 -- not spam, 3 -- is spam, 4 -- ignore
 #return unless $Con{$fh}->{maillog};
 if($parm == 4 && !$incomingOkMail) {
  d(364);
  delete $Con{$fh}->{maillogbuf};
  delete $Con{$fh}->{maillog};
  close $Con{$fh}->{maillogfh} if $Con{$fh}->{maillogfh};
  delete $Con{$fh}->{maillogfh};
  delete $Con{$fh}->{mailloglength};
 } elsif($parm > 1) {
  d(362);
  # we now know if it is spam or not -- open the file
  $text=$Con{$fh}->{maillogbuf}.$text;
  delete $Con{$fh}->{maillogbuf};
  my $fn=maillogFilename($parm-2,$text);
  $FH='FHaaaaa' unless $FH;
  if(open(++$FH,">$fn")) {
   #print "maillog open $FH -> $fn\n";
   $Con{$fh}->{maillogfh}=$FH;
   $Con{$fh}->{mailloglength}=0;
   binmode $FH;
  } else {
   mlog($fh, "error opening maillog '$fn': $!");
  }
 }
 if(my $h=$Con{$fh}->{maillogfh}) {
  print $h $text;
  if(($Con{$fh}->{mailloglength}+=length($text))>10000 || $text=~/(^|[\r\n])\.[\r\n]/) {
   d(366);
   close $h;
   delete $Con{$fh}->{maillog};
   delete $Con{$fh}->{maillogfh};
   delete $Con{$fh}->{mailloglength};
  }
 } elsif(length($Con{$fh}->{maillogbuf})<10000) {
  $Con{$fh}->{maillogbuf}.=$text;
 }
}


#####################################################################################
#                Web Configuration functions

sub webRequest {
 my ($fh,$head,$data)=@_;

 %webRequests=(
  '/stats' => \&ConfigStats,
  '/lists' => \&ConfigLists,
  '/analyze' => \&ConfigAnalyze,
  '/maillog' => \&ConfigMaillog,
 );

 ($page,$j,$qs)=/^\w+ ([^\? ]+)(\?(\S*))?/i;
 $qs=$data if defined $data;
 ($auth)=/Authorization: Basic (\S+)/i;
 ($user,$pass)=split(':',base64decode($auth));
 #print "p=$page>\nqs=$qs>\na=$auth>\n";
 $qs=~y/+/ /;
 (%qs)=split(/[=&]/,$qs);
 for $k (keys %qs) {$qs{$k}=~s/%([0-9a-fA-F][0-9a-fA-F])/pack(C,hex($1))/ge}
 if($pass eq $webAdminPassword) {
  ConfigQuit($fh) if($page eq '/quit');
  print $fh ((defined ($v=$webRequests{$page}))? $v->(): webConfig());
 } else {
  print $fh "HTTP/1.1 401 Unauthorized
WWW-Authenticate: Basic realm=\"Anti-Spam SMTP Proxy (ASSP) Configuration\"
Content-type: text/html

<html><body><h1>Unauthorized</h1>
</body></html>\n";
 }

 WebDone($fh);
}

sub ConfigQuit {
 mlog('',"Quit requested from admin interface");
 &SaveWhitelist;
 my $fh=shift;
 print $fh "HTTP/1.1 200 OK
Content-type: text/html

<html><body><h1>ASSP Terminated.</h1>
</body></html>
";
 exit;
}

sub ConfigStats {
 #mlog('',"Stats requested from admin interface");
 # TODO -- this should be meaningful instead of what it currently is...
 my $s;
 my $uptime=(time - $Stats{starttime})/(24*3600);
 my $mpd=sprintf("%.1f",$Stats{messages}/$uptime);
 my $spams=$Stats{spams}+$Stats{bspams};
 my $pct=sprintf("%.1f",$spams/($Stats{messages}-$Stats{locals}+1)*100);
 my $uptime=sprintf("%.3f",$uptime);
 my $tconns=$Stats{connects};
 if ($Stats{connects}==0) { $tconns=1000000 ; }
 my $rejpct=sprintf("%.1f%%",($Stats{norelays}/$tconns)*100);


 #for (sort keys %Stats) {  $s.="<tr><td align=right>$_</td><td>$Stats{$_}</td></tr>\n"; }
"HTTP/1.1 200 OK
Cache-control: no-cache
Content-type: text/html

<head><title>Anti-Spam SMTP Proxy (ASSP) Statistics</title>
<STYLE type=text/css>
<!-- a:hover { color: #0090E5 }
body,p,td {font-size: 11pt; font-family : Arial, sans-serif; }
H1, h2, h3 {font-variant: small-caps; }
//-->
</STYLE>
<body $Color4>
<p align=center><font size=+4><b><u>ASSP Statistics</u></b></font></p>
<p align=center>Back to <a href=/>configuration</a>.<p>
<table align=center border=1>
<tr><td colspan=2 align=center $Color3><font size=+2><b><i>SYSTEM Statistics</i></b></font></td></tr>
<tr><td align=right $Color1><b>ASSP Proxy Uptime:</b></td><td $Color2>$uptime days</td></tr>
<tr><td align=right $Color1><b>Total SMTP Connections received:</b></td><td $Color2>$Stats{connects}</td></tr>
<tr><td align=right $Color1><b>Total relay attempts rejected:</b></td><td $Color2>$Stats{norelays} ($rejpct)</td></tr>
<tr><td colspan=2 align=center $Color3><font size=+2><b><i>MESSAGE HANDLING Statistics</i></b></font></td></tr>
<tr><td align=right $Color1><b>Messages Processed:</b></td><td $Color2>$Stats{messages} ($mpd per day)</td></tr>
<tr><td align=right $Color1><b>Spams Rejected:</b></td><td $Color2>$spams</td></tr>
<tr><td align=right $Color1><b>Spams passed-thru due to 'SPAMLOVER' exception:</b></td><td $Color2>$Stats{spamlover}</td></tr>
<tr><td align=right $Color1><b>Local Mails:</b></td><td $Color2>$Stats{locals}</td></tr>
<tr><td align=right $Color1><b>Viruses Blocked:</b></td><td $Color2>$Stats{viri}</td></tr>
<tr><td align=right $Color1><b>Mail From Whitelisted Senders:</b></td><td $Color2>$Stats{whites}</td></tr>
<tr><td align=right $Color1><b>Percentage of non-local mail that is spam:</b></td><td $Color2>$pct%</td></tr>
</table>
<p align=center>ASSP Total Stats are <a href=http://assp.sourceforge.net/cgi-bin/total.pl target=_blank>here</a>.<p>
<p align=right>
<A href=http://sourceforge.net target=_blank>
<IMG src=http://sourceforge.net/sflogo.php?group_id=69172&amp;type=5
width=210 height=62 border=0 alt=\"SourceForge Logo\" hspace=2>
</p>
</A>
</body></html>
";
}

sub ConfigLists {
 #mlog('',"lists requested from admin interface");
 my $s;
 my $act=$qs{action};
 if($act) {
  my $color=$qs{list} eq 'red'? 'Red': 'White';
  my $list=$color."list";
  for ($qs{addresses}=~/([^<>,;"'\(\)\s]+\@[^<>,;"'\(\)\s\?]+)/ig) {
   $s.="<tr><td align=right>$_</td><td>";
   $_=lc $_;
   if($act eq 'v') {
    if($list->{$_}) {
     $s.="${color}listed";
    } else {
     $s.="<font color=red>NOT ${color}listed</font>";
    }
   } elsif($act eq 'a') {
    if($list->{$_}) {
     $s.="<font color=green>already ${color}listed</font>";
    } else {
     if($color eq 'White' && /\@($localDomains)$/i) {
      $s.="<font color=red>local addresses not allowed on whitelist</a>";
     } else {
      $s.="added";
      $list->{$_}=time;
      mlog(0,"${color}list addition: $_ (admin)");
     }
    }
   } elsif($act eq 'r') {
    if($list->{$_}) {
     $s.="removed";
     delete $list->{$_};
     mlog(0,"${color}list deletion: $_ (admin)");
    } else {
     $s.="not ${color}listed";
    }
   }
   $s.="</td></tr>\n";
  }
 }
 if($qs{B1}=~/^Show (.)/) {
  local $/="\n";
  if($1 eq 'R') {
   $RedlistObject->flush();
   open(F,"<$base/$redlistdb");
   $s.="<tr><td align=center><b>Redlist</b></td></tr>\n";
  } else {
   $WhitelistObject->flush();
   open(F,"<$base/$whitelistdb");
   $s.="<tr><td align=center><b>Whitelist</b></td></tr>\n";
  }
  my $l;
  while($l=<F>) {
   my ($a)=$l=~/([^\002]*)/;
   $s.="<tr><td>$a</td></tr>\n";
  }
  close F;
 }
<<EOT;
HTTP/1.1 200 OK
Cache-control: no-cache
Content-type: text/html

<head><title>Anti-Spam SMTP Proxy (ASSP) List Maintenance</title>
<STYLE type=text/css>
<!-- a:hover { color: #0090E5 }
body,p,td {font-size: 11pt; font-family : Arial, sans-serif; }
H1, h2, h3 {font-variant: small-caps; }
//-->
</STYLE>
<body $Color4><basefont face=ariel>
<p align=center><font size=+4 style=bold>
ASSP List Maintenance</font></p>
<p align=center>Back to <a href=/>configuration</a>.<p>
<table align=center border=1>$s</table>
<form method="POST">
    <div align="center"><center><table border="0" cellpadding="3">
        <tr>
            <td>Do you want to work with the:
            </td>
            <td>
            <input type=radio name=list value=white checked> Whitelist or<br>
            <input type=radio name=list value=red> Redlist
            </td>
        </tr>
        <tr>
            <td>Do you want to: </td>
            <td><input type="radio" name="action" value="a">add<br>
            <input type="radio" name="action" value="r">remove<br>
            <input type="radio" checked name="action" value="v">or verify</td>
            <td>List the addresses in this box:<br>
            <p><textarea name="addresses" rows="3" cols="40">$qs{addresses}</textarea></p>
            </td>
        </tr>
        <tr>
            <td>&nbsp;</td>
            <td><input type="submit" name="B1" value="Submit"></td>
            <td>&nbsp;</td>
        </tr>
    </table>
    </center></div><p>&nbsp;</p>
</form>
Post less than 1 megabyte of data at a time.<p>
Note: The redlist is not a blacklist. The redlist is a list of addresses that cannot
contribute to the whitelist, and who are not considered local, even if their mail is
from a local computer. For example, if someone goes on a vacation and turns on their
email's autoresponder, put them on the redlist until they return. Then as they reply
to every spam they receive they won't corrupt your non-spam collection or whitelist.

<form method=POST>
<table align=center>
<tr><td><input type=submit name=B1 value="Show Whitelist"></td>
<td><input type=submit name=B1 value="Show Redlist"></td></tr>
<tr><td colspan=2>Warning: If your whitelist or redlist is long, pushing these buttons
is ill-advised. Use these for testing and while your whitelist is short.</td></tr>
</table>
</form>

<p align=right>
<A href=http://sourceforge.net target=_blank>
<IMG src=http://sourceforge.net/sflogo.php?group_id=69172&amp;type=5
width=210 height=62 border=0 alt=\"SourceForge Logo\" hspace=2>
</p>
</A>
</body></html>
EOT
}

sub ConfigAnalyze {
 #mlog('',"lists requested from admin interface");
 my ($s,$st,$wl,%wl);
 my $mail=$qs{mail};
 if($mail) {
  my $name=$myName; $name=~s/(\W)/\\$1/g;
  my ($ip)=$mail=~/Received: from ([0-9\.]+) .* by $name/i;
  while($mail=~/([^<>,;"'\(\)\s]+\@[^<>,;"'\(\)\s\?]+)/g) {
   my $a=lc $1;
   next if $wl{$a}++;
   if( $Redlist{$a} ){
    $wl.="<b>$1 is redlisted<br></b>\n"
   } else {
    $wl.="<b>$1 is whitelisted<br></b>\n" if $Whitelist{$a};
   }
  }
  $wl.="<p>" if $wl;
  $wl.="<b>Mail matches White RE: '$&'</b><p>\n" if $whiteRe && $mail=~/$whiteRe/is;
  $wl.="<b>Mail matches Black RE: '$&'</b><p>\n" if $blackRe && $mail=~/$blackRe/is;

  $mail=clean(substr($mail,0,10000));
  $wl.="<b>Mail matches White RE: '$&'</b><p>\n" if $whiteRe && $mail=~/$whiteRe/is;
  $wl.="<b>Mail matches Black RE: '$&'</b><p>\n" if $blackRe && $mail=~/$blackRe/is;

  my @t; my %got;

  if(defined($Dnsbl{$ip}) || defined($Dnsbl{$ip3})) {
   $wl.="<b>$ip dnsbl hit</b> (adds 0.97 0.97)<p>\n";
   push(@t,0.97);
   push(@t,0.97);
  }
  if(!$NoGoodhosts && ! $Goodhosts{$ip}) {
   my $p=$Goodhosts{'spam'}/($Goodhosts{'notspam'}+$Goodhosts{'spam'}+1);
   $p=sprintf("%.4f",$p);
   if($p > 0.65) {
    $wl.="<b>$ip is not on the goodhost list</b> (adds $p $p)<p>\n";
    push(@t,$p);
    push(@t,$p);
   } else {
    $wl.="$ip is not on the goodhost list but your ratio is still too small: $p ($Goodhosts{notspam} to $Goodhosts{spam}<p>";
   }
  }

  my ($v,$lt,$llt,$t,%seen);
  while($mail=~/([-\$A-Za-z0-9\'\.\,!]+)/g) {
   next if length($1) > 20;
   $llt=$lt; $lt=$t; $t=lc $1;
   $t=~s/[,.']+$//; $t=~s/!!!+/!!/g; $t=~s/--+/-/g;
   my $j="$llt $lt $t";
   next if $seen{$j}++>1;
   push(@t,$v) if ($v=$Spamdb{$j}); $got{$j}=$v if $v;
  }
  my $cnt=0;
  $s.="<tr><td>Bad Words</td><td>Bad Prob</td><td>Good Words</td><td>Good Prob</td></tr>\n";
  for (sort {abs($got{$b}-.5)<=>abs($got{$a}-.5)} keys %got) {
   my $g=sprintf("%.4f",$got{$_});
   if($g < 0.5) {
    $s.="<tr><td></td><td></td><td align=right>$_</td><td>$g</td></tr>\n";
   } else {
    $s.="<tr><td align=right>$_</td><td>$g</td><td></td><td></td></tr>\n";
   }
   last if $cnt++ > 20;
  }
  @t=sort {abs($b-.5)<=>abs($a-.5)} @t;
  @t=@t[0..30];
  $st="<br>Analysis totals: ";
  for (@t) {$st.=sprintf ("%.4f ",$_) if $_;}
  $st.="<p>\n";
  #print "\n"; ,join(',',@t),"\n";
  my $p1=1; my $p2=1; for $p (@t) {if($p) {$p1*=$p; $p2*=(1-$p);}}
  $p1=$p1/($p1+$p2);
  $st.=sprintf("spam-prob = %.5f<p>\n",$p1);
  $mail=~s/([^\n]{70,84}[^\w\n<\@])/$1\n/g; $mail=~s/\n+/\n/g;
 }
<<EOT;
HTTP/1.1 200 OK
Cache-control: no-cache
Content-type: text/html

<head><title>Anti-Spam SMTP Proxy (ASSP) Mail Analyzer</title>
<STYLE type=text/css>
<!-- a:hover { color: #0090E5 }
body,p,td {font-size: 11pt; font-family : Arial, sans-serif; }
H1, h2, h3 {font-variant: small-caps; }
//-->
</STYLE>
<body $Color4><basefont face=ariel>
<p align=center><font size=+4 style=bold>
ASSP Mail Analyzer</font></p>
<p align=center>Back to <a href=/>configuration</a>.<p>
<p>This page will show you how ASSP analyzes an email to come up with the assigned spam
probability. You can also see how it pre-processes mail.</p>
$wl<table align=center border=1>$s</table>$st
<form method="POST">
    <table border="0" cellpadding="3" align=center>
        <tr>
            <td align="center">Copy and paste the mail header and
            body here:<br>
            <textarea name="mail" rows="10" cols="85">$mail</textarea>
            </td>
        </tr>
        <tr>
            <td align="center"><input type="submit" name="B1"
            value="Analyze"></td>
        </tr>
    </table>
</form>
<p>Note: Analysis is performed using the current spam database --
if yours was rebuilt since the time the mail was received you'll
receive a different result.</p>

<p>To use this form using <i>Outlook Express</i> do the following. Right-click on the message
of interest. Select <i>Properties</i>. Click the <i>Details</i> tab. Click the <i>message
source</i> button. Right-click on the message source and click <i>Select All</i>. Right-click
again and click <i>Copy</i>. Click on the text box above and paste (Ctrl-V perhaps). Click
the <i>Analyze</i> button.</p>
<p>The page will update to show you the following: if any of the email's addresses are in
the redlist or whitelist, the most and least spammy phrases together with their spaminess,
the resulting probabilities (probabilities may repeat one time), and the final spam probability
score.</p>
<p align=right>
<A href=http://sourceforge.net target=_blank>
<IMG src=http://sourceforge.net/sflogo.php?group_id=69172&amp;type=5
width=210 height=62 border=0 alt=\"SourceForge Logo\" hspace=2>
</p>
</A>
</body></html>
EOT
}

sub ConfigMaillog {
 open(F,"<$base/$logfile");
 #print "$base/$logfile e=$!\n";
 seek(F,-10000,2);
 local $/;
 my $s=<F>;
 close F;
 $s=~s/&/&amp;/g;
 $s=~s/</&lt;/g;
 $s=~s/>/&gt;/g;
<<EOT;
HTTP/1.1 200 OK
Cache-control: no-cache
Content-type: text/html

<head><title>Anti-Spam SMTP Proxy (ASSP) Maillog Tail</title>
<STYLE type=text/css>
<!-- a:hover { color: #0090E5 }
body,p,td {font-size: 11pt; font-family : Arial, sans-serif; }
H1, h2, h3 {font-variant: small-caps; }
//-->
</STYLE>
<body $Color4><basefont face=ariel>
<p align=center><font size=+4 style=bold>
ASSP Maillog Tail</font></p>
<p align=center>Back to <a href=/>configuration</a>.<p>
<p>Press your browser's refresh to update this screen. Newest entries are at the end.</p><hr>
<pre>$s</pre>
<hr>
<p align=right>
<A href=http://sourceforge.net target=_blank>
<IMG src=http://sourceforge.net/sflogo.php?group_id=69172&amp;type=5
width=210 height=62 border=0 alt=\"SourceForge Logo\" hspace=2>
</p>
</A>
</body></html>
EOT
}

sub webConfig {
 my $r="HTTP/1.1 200 OK
Content-type: text/html
Cache-control: no-cache

<html>
<head><title>Anti-Spam SMTP Proxy (ASSP) Configuration</title>
<STYLE type=text/css>
<!-- a:hover { color: #0090E5 }
body,p,td {font-size: 11pt; font-family : Arial, sans-serif; }
H1, h2, h3 {font-variant: small-caps; }
//-->
</STYLE>
<body $Color4><basefont face=ariel>
<table width=100%><tr>
<td width=176>
<a href=http://assp.sourceforge.net/ target=_blank>
<img src=http://assp.sourceforge.net/assp.gif alt=\"SPAM is Yummy!\" align=right hspace=2 border=0
width=174 height=115></a>
</td>
<td align=center>
<font size=+4 style=bold>
<a href=http://assp.sourceforge.net/ target=_blank>ASSP</a> Configuration</font><br>
Your version: $version<br>
Latest Documentation is <a href=http://assp.sourceforge.net/fom/cache/1.html target=_blank>here</a>.<br>
<a href=stats>Statistics</a>,
<a href=lists>Update / Verify the Whitelist or Redlist</a>,
<a href=analyze>Analyze an email</a>,
<a href=maillog>See the maillog tail</a>
</td>
</tr></table>
Note that anywhere | is used that if you put two together (||) you will get an undesirable result. Don't do that.<p>
<form method=post><table bgcolor=#FFFFFF>\n";
 $ConfigChanged=0;
 for $c (@Config) {$r.= "<tr>".$c->[3]->(@{$c})."</tr>\n";}
 if($ConfigChanged) {
  SaveConfig();
 }
 my $quit="<p align=left><form method=post action=quit>Panic button: <input type=submit value=\"Terminate ASSP now!\"></form></p>" unless $AsAService;
 $r.= "
<tr><td $Color1>&nbsp;</td><td align=center $Color2><input type=submit value=\"Apply Changes\"></td></tr>
</table>
</form>

$quit
<p align=right>
<A href=http://sourceforge.net target=_blank>
<IMG src=http://sourceforge.net/sflogo.php?group_id=69172&amp;type=5
width=210 height=62 border=0 alt=\"SourceForge Logo\" hspace=2>
</p>
</A>
</body></html>\n";
 $r;
}

sub SaveConfig {
 rename("$base/assp.cfg.bak","$base/assp.cfg.bak.bak");
 rename("$base/assp.cfg","$base/assp.cfg.bak");
 open(F,">$base/assp.cfg");
 for(sort keys %Config) {print F "$_:=$Config{$_}\n";}
 close F;
}

sub textinput {my ($name,$nicename,$size,$func,$default,$valid,$onchange,$description)=@_;
 my $Error=checkUpdate($name,$valid,$onchange);
 "<td align=right $Color1>$nicename</td><td $Color2><input name=$name size=$size value=\"$Config{$name}\"><br>\n$Error$description</td>\n";
}
# everybody wants this, but I hate it -- use it if you care.
sub passinput {my ($name,$nicename,$size,$func,$default,$valid,$onchange,$description)=@_;
 my $Error=checkUpdate($name,$valid,$onchange);
 "<td align=right $Color1>$nicename</td><td $Color2><input type=PASSWORD name=$name size=$size value=\"$Config{$name}\"><br>\n$Error$description</td>\n";
}
sub checkbox {my ($name,$nicename,$size,$func,$default,$valid,$onchange,$description)=@_;
 my $Error=checkUpdate($name,$valid,$onchange);
 my $checked=$Config{$name}?'checked':'';
 "<td $Color1>&nbsp;</td><td $Color2><input type=checkbox name=$name value=1 $checked> $nicename<br>\n$Error$description</td>\n";
}
sub heading {my ($description)=$_[4];
 "<td colspan=2 align=center $Color3><font size=+2>$description</font></td>";
}

sub checkUpdate {
 my ($name,$valid,$onchange)=@_;
 return '' unless %qs;
 if($qs{$name} ne $Config{$name}) {
  if($qs{$name}=~/$valid/i) {
   my $new=$1; my $info;
   my $old=$Config{$name};
   $Config{$name}=$new;
   if($onchange) {
    $info=$onchange->($name,$old,$new);
   } else {
    mlog(0,"Admin update: $name changed from '$old' to '$new'");
    ${$name}=$new;
    # -- this sets the variable name with the same name as the config key to the new value
    # -- for example $Config{myName}="ASSP-nospam" -> $myName="ASSP-nospam";
   }
   $ConfigChanged=1;
   return "<font color=green><b>*** Updated $info</b></font><br>";
  } else {
   return "<font color=red><b>*** Invalid: '$qs{$name}'</b></font><br>";
  }
 }
}

sub fixConfigSettings {
 $Config{'base'}=$base;
    # -- this sets the variable name with the same name as the config key to the new value
    # -- for example $Config{myName}="ASSP-nospam" -> $myName="ASSP-nospam";
 for (keys %Config) {${$_}=$Config{$_};}

 $EmailFrom="<>" unless $EmailFrom;
  # turn settings into regular expressions
 for (@Config) {
  if($_->[6] eq 'ConfigMakeRe') {
   ${$_->[0]}=~s/([\.\[\]\-\(\)\*\+\\])/\\$1/g;
  }
 }
}

sub ConfigMakeRe { my ($name, $old, $new)=@_;
 $$name=$new;
 mlog(0,"Admin update: $name changed from '$old' to '$new'");
 $$name=~s/([\.\[\]\-\(\)\*\+\\])/\\$1/g;
 '';
}

# convert a multiline string to a regular expression
sub re {
 $_=$_[0];
 s/\n/\|/g;
 s/\|$//;
 $_;
}

sub ConfigChangeMailPort {my ($name, $old, $new)=@_;
 if($> == 0 || $new >= 1024) {
  # change the listenport
  $listenPort=$new;
  if($lsn) {
   $readable->remove($lsn);
   $lsn->close();
  }
  $lsn = newListen($listenPort,\&NewSMTPConnection);
  mlog(0,"Listening on new mail port $listenPort (changed from $old) per admin request");
  return '';
 } else {
  # don't have permissions to change
  mlog(0,"Request to listen on new mail port $listenPort (changed from $old) -- restart required; euid=$>");
  return "<br>Restart required; euid=$>";
 }
}

sub ConfigChangeAdminPort {my ($name, $old, $new)=@_;
 if($> == 0 || $new >= 1024) {
  # change the listenport
  $webAdminPort=$new;
  $readable->remove($WebSocket);
  $WebSocket->close();
  $WebSocket = newListen($webAdminPort,\&NewWebConnection);
  if($WebSocket) {
   mlog(0,"Listening on new admin port $new (changed from $old) per admin request");
  } else {
   # couldn't open the port -- switch back
   $webAdminPort=$old;
   $WebSocket = newListen($webAdminPort,\&NewWebConnection);
   mlog(0,"Couldn't open new port -- still listening on $old");
   $Config{$name}=$old;
   return "<font color=red>Couldn't open new port $new</font>";
  }
  return '';
 } else {
  # don't have permissions to change
  mlog(0,"Request to listen on new admin port $new (changed from $old) -- restart required; euid=$>");
  return "<br>Restart required; euid=$>";
 }
}

sub ConfigChangeRelayPort {my ($name, $old, $new)=@_;
 unless ($relayHost && $new) {
  if($Relay) {
   $readable->remove($Relay);
   $Relay->close();
   mlog(0,"Relay port disabled");
   return '<br>relay port disabled';
  } else {
   return "<br>relayHost ($relayHost) and relayPort ($new) must be defined to enable relaying";
  }
 }
 if($> == 0 || $new >= 1024) {
  # change the listenport
  $relayPort=$new;
  if($Relay) {
   $readable->remove($Relay);
   $Relay->close();
  }
  $Relay=newListen($relayPort,\&NewSMTPConnection);
  mlog(0,"Listening for relay connections at $relayPort -- changed per admin request");
  return '';
 } else {
  # don't have permissions to change
  mlog(0,"Request to listen on new relay port $new (changed from $old) -- restart required; euid=$>");
  return "<br>Restart required; euid=$>";
 }
}

sub ConfigChangeLogfile {my ($name, $old, $new)=@_;
 close LOG if $logfile;
 $logfile=$new;
 if($logfile && open(LOG,">>$base/$logfile")) {my $oldfh = select(LOG); $| = 1; select($oldfh);}
 mlog(0,"Log file changed to '$new' from '$old' per admin request");
 '';
}
sub ConfigDEBUG {my ($name, $old, $new)=@_;
 close DEBUG if $DEBUG;
 $DEBUG=$new;
 if($DEBUG) {
  open(DEBUG, ">$base/".time.".dbg"); binmode(DEBUG); my $oldfh = select(DEBUG); $| = 1; select($oldfh);
  eval(q[sub d {
   $time=gmtime(); $time=~s/... (...) (..) (........) ..(..)/$2 $1 $4 $3/;
   print DEBUG "$time <$_[0]>";
  }
  ]);
 } else {
  eval(q[sub d{}]);
 }
 mlog(0,"Deubg file changed to '$new' from '$old' per admin request");
 '';
}

sub reloadConfigFile {
 # called on SIG HUP
 my %newConfig;
 mlog(0,"Sig HUP -- reloading config");
 open(F,"<$base/assp.cfg"); local $/; (%newConfig)=split(/:=|\n/,<F>); close F;
 for $c (@Config) {
  my ($name,$nicename,$size,$func,$default,$valid,$onchange,$description)=@$c;
  if($Config{$name} ne $newConfig{$name}) {
   if($newConfig{$name}=~/$valid/i) {
    my $new=$1; my $info;
    if($onchange) {
     $info=$onchange->($name,$Config{$name},$new);
    } else {
     mlog(0,"Admin update: $name changed from '$Config{$name}' to '$new'");
     ${$name}=$new;
     # -- this sets the variable name with the same name as the config key to the new value
     # -- for example $Config{myName}="ASSP-nospam" -> $myName="ASSP-nospam";
    }
    $Config{$name}=$new;
   } else {
    mlog(0, "Error: invalid '$newConfig{$name}' -- not changed");
   }
  }
 }
}

#####################################################################################
#                orderedtie
{
package orderedtie;
# This is a tied value that caches lookups from a sorted file; \n seperates records,
# \002 seperates the key from the value. After n (currently 2000) lookups the cache is cleared.
# This give us most of the speed of the hash without the huge memory overhead of storing the
# entire hash and should be totally portable. Picking the best value for n requires some tuning.
# A \n is required to start the file.

# if you're updating entries it behoves you to call flush every so often to make sure that your
# changes are saved. This also frees the memory used to remember updated values.

# for my purposes a value of undef and a nonexistant key are the same

# Obviosly if your keys or values contain \n or \002 it will totally goof things up.


sub TIEHASH {
 my ($c,$fn)=@_;
 my $self={
  fn => $fn,
  age => mtime($fn),
  cnt => 0,
  cache => {},
  updated => {},
  ptr => 1,
 };
 bless $self, $c;
 return $self;
}
sub DESTROY { $_[0]->flush(); }

sub mtime { my @s=stat($_[0]); $s[9]; }

sub flush {
 my $this=shift;
 return unless %{$this->{updated}};
 my $f=$this->{fn};
 #print "flushing $f\n";
 open(O,">$f.tmp") || return undef;
 binmode(O);
 open(I,"<$f") || print O "\n";
 binmode(I);
 local $/="\n";
 my @l=(sort keys %{$this->{updated}});
 my ($k,$d,$r,$v);
 while($r=<I>) {
  ($k,$d)=split("\002",$r);
  while(@l && $l[0] lt $k) {
   $v=$this->{updated}{$l[0]};
   print O "$l[0]\002$v\n" if defined $v;
   shift(@l);
  }
  if($l[0] eq $k) {
   $v=$this->{updated}{$l[0]};
   print O "$l[0]\002$v\n" if defined $v;
   shift(@l);
  } else {
   print O $r;
  }
 }
 while(@l) {
  $v=$this->{updated}{$l[0]};
  print O "$l[0]\002$v\n" if defined $v;
  shift(@l);
 }
 close I; close O; rename("$f.tmp", $f);
 $this->{updated}={};
}

sub STORE {
 my ($this, $key, $value)=@_;
 $this->{cache}{$key}=$this->{updated}{$key}=$value;
}

sub FETCH { my ($this, $key)=@_;
 return $this->{cache}{$key} if exists $this->{cache}{$key};
 $this->resetCache() if($this->{cnt}++ >5000 || ($this->{cnt} & 0x1ff) == 0 && mtime($this->{fn}) != $this->{age});
 return $this->{cache}{$key}=binsearch($this->{fn},$key);
}

sub resetCache {
 my $this=shift;
 $this->{cnt}=0;
 $this->{age}=mtime($this->{fn});
 $this->{cache}=\%{$this->{updated}};
 main::mlog(0,"cache reset ($this->{fn})");
}

sub binsearch {
 my ($f,$k)=@_;
 open(F,"<$f") || return undef;
 binmode(F);
 my $siz=my $h=-s $f;
 $siz-=1024;
 my $l=0;
 my $k0=$k;
 $k=~s/([\[\]\(\)\*\^\!\|\+\.\\\/\?\`\$\@\{\}])/\\$1/g; # make sure there's no re chars unqutoed in the key
 #print "k=$k ($_[1])\n";
 while($h > $l ) {
  my $m=(($l+$h)>>1)-1024;
  $m=0 if $m < 0;
  #print "($l $m $h) ";
  seek(F,$m,0);
  my $d; my $read= read(F,$d,2048);
  if( $d=~/\n$k\002([^\n]*)\n/) {
   close F;
   #print "got $1\n";
   return $1;
  }
  my ($pre,$first,$last,$post)=$d=~/^(.*?)\n(.*?)\002.*\n(.*?)\002.*?\n(.*?)$/s;
  if($k0 gt $first && $k0 lt $last) {
   close F;
   #print "got miss\n";
   return undef;
  }
  if($k0 lt $first) {
   last if $m ==0;
   $h=$m-1024+length($pre);
  }
  if($k0 gt $last) {
   last if $m >= $siz;
   $l=$m+$read-length($post);
  }
 }
 close F;
 return undef;
}

sub FIRSTKEY { $this=shift;
 $this->flush();
 $this->{ptr}=1;
 $this->NEXTKEY();
}
sub NEXTKEY { my ($this, $lastkey)=@_;
 local $/="\n";
 open(F,"<$this->{fn}") || return undef;
 binmode(F);
 seek(F,$this->{ptr},0);
 my $r=<F>;
 return undef unless $r;
 $this->{ptr}=tell F;
 close F;
 my ($k,$v)=$r=~/(.*?)\002(.*?)\n/s;
 if(!exists($this->{cache}{$k}) && $this->{cnt}++ >2000) {
  $this->{cnt}=0;
  $this->{cache}=\%{$this->{updated}};
 }
 $this->{cache}{$k}=$v;
 $k
}

sub EXISTS { my ($this, $key)=@_;
 return defined FETCH($this, $key);
}

sub DELETE {my ($this, $key)=@_;
 $this->{cache}{$key}=$this->{updated}{$key}=undef;
}

sub CLEAR {my ($this)=@_;
 open(F,">$this->{fn}"); binmode(F); print "\n"; close F;
 $this->{cache}={};
 $this->{updated}={};
 $this->{cnt}=0;
}
}

