#!/usr/bin/perl

# rebuilds bayesian spam database
# (c) John Hanna 2003 under the terms of the GPL

open(F,"<assp.cfg"); local $/; (%Config)=split(/:=|\n/,<F>); close F;
$spamlog=$Config{spamlog} && "$Config{base}/$Config{spamlog}" || 'spam';
$notspamlog=$Config{notspamlog} && "$Config{base}/$Config{notspamlog}" || 'notspam';
$correctedspam=$Config{correctedspam} && "$Config{base}/$Config{correctedspam}" || 'errors/spam';
$correctednotspam=$Config{correctednotspam} && "$Config{base}/$Config{correctednotspam}" || 'errors/notspam';
$spamdb=$Config{spamdb} && "$Config{base}/$Config{spamdb}" || 'spamdb';
$MaxWhitelistDays=$Config{MaxWhitelistDays} || 400;

use bytes; # get rid of anoying 'Malformed UTF-8' messages

getmaxtick('rebuild');

# add(isspam,filename,weight)
# read the corrections
add(1,$_,2,\&spamHash) for (glob("$correctedspam/*"));
print "\n\n";

add(0,$_,4,\&hamHash) for (glob("$correctednotspam/*"));
print "\n\n";

# read spam database
add(1,$_,1,\&checkspam) for (glob("$spamlog/*"));

print "\n\n";
# read non-spam database
add(0,$_,1,\&checkham) for (glob("$notspamlog/*"));

print "\n\n";

# update stats
#$norm=($SpamWordCount +$HamWordCount)/(2*$SpamWordCount) if $SpamWordCount;
#$norm=1 if $norm > 1;
print "swc=$SpamWordCount hwc=$HamWordCount, norm=$norm \n";
for(keys %tot) {
 next if $tot{$_} < 5;
 ($s,$t)=($spam{$_},$tot{$_});
 # if token represents all spam or all ham then square its value
 if($s==$t || $s==0) {
  $s=$s*$s; $t=$t*$t;
 }
 $v=(1+$s)/($t+2);
 $spa{$_}=$v if abs($v-.5) > .05;
 &tick if ($Tick++ & 0x3f)==0;
}

# done with these -- free some ram
undef %spam;
undef %tot;

 open(F,">$spamdb.tmp") || die "couldn't open $spamdb.tmp: $!\n";
 binmode F;
 print F "\n";
 for $key (sort keys %spa) {
  &tick if ($Tick++ & 0x3f)==0;
  print F "$key\002$spa{$key}\n";
 }
 close F;
 rename($spamdb,"$spamdb.bak");
 rename("$spamdb.tmp",$spamdb);

 if(rand()< .05) {
  # rarely, let's clean the whitelist of old entries
  $t=time - 24*3600*$MaxWhitelistDays;
  print "Cleaning whitelist\n";

  open(F,"<$whitelistdb");
  binmode(F);
  open(O,">$whitelistdb.tmp");
  binmode(O);
  local $/="\n";
  <F>; print O "\n";
  while(<F>) {
   my ($a,$t)=split("\002",$_);
   next if $t > $time;
   print O;
  }
  close F; close O;
  rename($whitelistdb,"$whitelistdb.bak");
  rename("$whitelistdb.tmp",$whitelistdb);

  open(F,"<goodhosts");
  binmode(F);
  open(O,">goodhosts.tmp");
  binmode(O);
  $t=time - 24*3600*20;
  local $/="\n";
  <F>; print O "\n";
  while(<F>) {
   my ($a,$t)=split("\002",$_);
   next if $t > $time;
   print O;
  }
  close F; close O;
  rename('goodnosts',"goodhosts.bak");
  rename("goodhosts.tmp",'goodhosts');
 }

putmaxtick('rebuild');

sub spamHash { $SpamHash{hash($_[1])}=''; }

sub hamHash { $HamHash{hash($_[1])}=''; }

sub checkspam {
 my $h;
 if(defined($HamHash{$h=hash($_[1])}) ) {
  # we've found a message in the spam database that is the same as one in the corrected Ham group
  my $fn=shift;
  # delete it
  #print "$fn is spam match\n'$h' -> $HT{$h}\n";
  #$fn2=$fn; $fn2=~s/spam/spam2/;
  #rename($fn,$fn2);
  unlink($fn);
  return 1;
 }
 0;
}

sub checkham {
 my $h;
 if(defined($SpamHash{$h=hash($_[1])}) ) {
  # we've found a message in the ham database that is the same as one in the corrected spam group
  my $fn=shift;
  # delete it
  #print "$fn is hpam match\n'$h' -> $HT{$h}\n";
  #$fn2=$fn; $fn2=~s/spam/spam2/;
  #rename($fn,$fn2);
  unlink($fn);
  return 1;
 }
 0;
}

sub get {
 my ($fn,$sub)=@_;
 open(F,"<$fn") || return '';
 my $m;
 read(F,$m,10000); $orig=$m;
 close F;
 return '' if $sub->($fn,$m);
 clean($m);
}

sub add {
 #print "+" if $Counter++ % 100 ==0;
 &tick if ($Tick++ & 0x3f)==0;
 my ($spam,$fn,$factor,$sub)=@_;
 #print "$fn <$spam> [$factor]\n";
 return if -d $fn;
 my ($t,$lt,$llt);
 local $_=get($fn,$sub);
 return unless $_; # use $$sub to identify and remove spam or ham that matches corrected items
 while(/([-\$A-Za-z0-9\'\.\,!]+)/g) {
  next if length($1) > 20 || length($1) < 2;
  $llt=$lt; $lt=$t; $t=lc $1;
  #next if $t=~/^\d/; # ignore numbers
  $t=~s/[,.']+$//; $t=~s/!!!+/!!/g; $t=~s/--+/-/g;
  if($spam) {$SpamWordCount+=$factor} else {$HamWordCount+=$factor}
  #$spam{$t}+=$factor if $spam;
  #$tot{$t}+=$factor;
  #$spam{"$lt $t"}+=$factor if $spam;
  #$tot{"$lt $t"}+=$factor;
  $spam{"$llt $lt $t"}+=$factor if $spam;
  $tot{"$llt $lt $t"}+=$factor;
 }
}

# clean up source email
sub clean {
 local $_="\n".shift;
 # replace &#ddd encoding
 s/&#(\d{1,3});?/chr($1)/ge;
 #s/base64.{0,99}\n\n([a-zA-Z0-9+\/\n=]+)/base64decode($1)/gse;
 # replace base64 encoding
 s/\n([a-zA-Z0-9+\/=]{40,}\r?\n[a-zA-Z0-9+\/=\r\n]+)/base64decode($1)/gse;
 # clean up quoted-printable references
 if(/quoted-printable/) {
  s/=\r?\n//g;
  s/=([0-9a-fA-F]{2})/pack("C",hex($1))/gei;
 }
 s/(http:\/\/\S+)/fixurl($1)/ige;
 # strip out mime continuation
 s/.*---=_NextPart_.*\n//g;
 # mark the subject
 s/\nsubject: (.*)/fixsub($1)/ige;
 # remove received lines
 s/\n(received|Content-Type): .*(\n[\t ].*)*//ig;
 # remove other header lines
 s/\n[a-zA-Z\-]{2,40}: .*//g;
 # clean up &nbsp; and &amp;
 s/&nbsp;?/ /gi; s/&amp;?/and/gi;
 # clean up html stuff
 while(s/(\w+)(<[^>]*>)((<[^>]*>)*\w+)/$2$1$3/g){} # move html out of words
 s/<([biu]|strong)>/ boldifytext /gi; s/<\/?(p|br|div|t[dr])[^>]*>/\n/gi; s/<\/([biua]|font|strong)>//gi;
 s/<\/?(html|meta|head|body|span|o)[^>]*>//ig;
 # treat titles like subjects
 s/<title[^>]*>(.*?)<\/title>/fixsub($1)/ige;
 # remove style sheets
 s/<style[^>]*>.*?<\/style>//igs;
 # remove html comments
 s/<!.*?-->//gs; s/<![^>]*>//g;
 # look for random words
 s/ [bcdfghjklmnpqrstvwxz]{3}[bcdfghjklmnpqrstvwxz0-9]{3,}/ randword randword randword /gi;
 # look for linked images
 s/(<a[^>]*>[^<]*<img)/ linkedimage linkedimage linkedimage $1/gi;
 #open(F,">t"); print F $_; close F;
 $_;
}

sub fixurl {
 my $a=shift;
 $a=~s/%([0-9a-fA-F][0-9a-fA-F])/pack('C',hex($1))/ge;
 $a;
}

sub fixsub {
 my $s=shift;
 #print "$s=>";
 $s=~s/ {3,}/ lotsaspaces /g;
 $s=~s/(\S+)/ssub $1/g;
 #print "$s\n";
 "\nssub $s ssub";
}

sub base64decode {
 my $str = shift;
 my $res = "\n\n";
 $str =~ tr|A-Za-z0-9+/||cd;
 $str =~ tr|A-Za-z0-9+/| -_|;
 while ($str =~ /(.{1,60})/gs) {
  my $len = chr(32 + length($1)*3/4);
 $res .= unpack("u", $len . $1 );
 }
 $res;
}

sub hash {
 # creates a $len length hash of $msg
 my $msg=shift;
 my $len=20;
 my ($head,$bod)=$msg=~/^(.*?)\n\r?\n(.*)/s;
 my ($sub)=$msg=~/Subject: (.*)/i;
 my $hash=substr($sub,0,20).substr($bod,0,260);
 my @a=(0 .. 94);
 my $c=0;
 for my $n (0 .. length($hash)) {
  $c=($a[$c]+$c+ord(substr($hash,$n,1))) % 95;
  my $n2=$n % 95;
  @a[$n2,$c]=@a[$c,$n2];
 }
 for my $n (0 .. length($hash)) {
  $c=($a[$c]+$c+ord(substr($hash,$n,1))) % 95;
  my $n2=$n % 95;
  @a[$n2,$c]=@a[$c,$n2];
 }
 my $r='';
 $c=0;
 for my $n (@a) {
  $r.=chr($n+32);
  last if ++$c > $len;
 }
 #$HT{$r}=$hash;
 $r;
}

sub tick {
  my $stars=(70 * $Tick / $MaxTick); $stars=70 if $stars > 70;
  $stars='*' x $stars;
  print "$Tick $stars \r";
}
sub getmaxtick {
  if(open(F,"<$_[0].mt")) {
   $MaxTick = <F>;
   close F;
   $MaxTick=~y/0-9//cd;
   $MaxTick+=0;
  }
  $MaxTick=1000000 unless $MaxTick>1000;
  print "mt=$MaxTick\n";
}
sub putmaxtick {
  open(F,">$_[0].mt");
  print F $Tick;
  close F;
}
